"""Модуль для генерации признаков."""

import pandas as pd
import numpy as np
from abc import ABC, abstractmethod
import itertools

from astropy.coordinates import SkyCoord
import astropy.units as u

from typing import Union


class Transform(ABC):
    """
    Abstract class for transforms.

    You can realize the heir of the class by implementing two methods:

    TODO.
    """

    @abstractmethod
    def __init__(self, *args: list, **kwargs: dict) -> None:
        """Init."""
        self.in_columns: dict[str, str] = {}
        self.before_transforms: dict[tuple[str, ...], Transform] = dict()
        self.out_columns: list[str] = []
        # You code here
        pass

    # Optional
    def optimize(self, df: pd.DataFrame) -> None:
        """
        Возможный для переопределения метод.

        Метод вызывается перед вызовом transform. В нем ножно указать, какие колонки не нужно перегенерировать,\
              если они уже есть во входном DataFrame.
        """
        return None

    @abstractmethod
    def _transform(self, df: pd.DataFrame) -> None:
        """Метод для переопределения. Производит трансформацию."""
        pass

    def __str__(self) -> str:
        """Str."""
        return self.__class__.__name__

    def __repr__(self) -> str:
        """Str."""
        return self.__class__.__name__

    def untransform(self, df: pd.DataFrame) -> None:
        """Undo transformation."""
        for col in self.out_columns:
            if col in df.columns:
                del df[col]

    def _make_before_transforms(self, df: pd.DataFrame, columns: dict[str, str]) -> list[str]:
        """Производит трансформации, которые генерируют колонки, необходимые для выполнения метода transform."""
        start_columns = df.columns
        col_to_del = []
        for cols, transform in self.before_transforms.items():
            old_columns = df.columns
            transform.transform(df, columns)
            assert set(df.columns) == set(list(old_columns) + list(cols))
            for col in cols:
                if col not in start_columns:
                    col_to_del.append(col)
        return col_to_del

    def get_columns(self) -> dict[str, str]:
        """
        Возвращает необходимые колонки с данными в виде словаря: ключ - новое значение, нужное внутри класса,\
            значение - описание колонки, которое необходимо переопределить названием колонки в вашем датасете.

        Словарь с переопределенными значениями нужно передать в метод transform.
        """
        columns = self.in_columns
        for cols, transform in self.before_transforms.items():
            columns.update(transform.get_columns())
        return columns

    def _check_columns(self, df: pd.DataFrame, columns: dict[str, str]) -> None:
        """
        Проверка, есть ли все нужны колонки с данными в датасете. Готов ли он к переименованию колонок.

        Args:
            df: Набор данных, готовый к переименовыванию колонок.
            columns: Словарь с ключем - старое название колонки, значением - новое название колонки.
        Raises:
            KeyError: Колонка данных должна находиться в наборе данных.
        """
        for i in columns.keys():
            assert i in df.columns, f'Колонка {i} не найдена в наборе данных.'

    def transform(self, df: pd.DataFrame, columns: Union[dict[str, str], None] = None,
                  ignore_optimize: bool = False) -> None:
        """
        Производит преобразование каталога.

        Args:
            df: Каталог с набором данных для преобразования.
            name_of_columns: Словарь с наименованиями колонок, который будет использоваться для преобразования внешних имён во внутренние.\
                  Подробнее в get_name_of_columns.
            ignore_optimize: При значении True - все уже существующие колонки будут переопределяться.\
                  В случае False - если колонка с соответствующим название есть внутри набора, то она не будет изменяться.\
                  Для работы необходимо переопределние метода optimize внутри наследуемого класса.
        Note:
            Входной каталог изменяется - добавляются новые колонки данных.
        """
        # TODO fill columns
        # Если колонки не заданы, то считаем, что переименовывать не надо. Делает тривиальное переименование.
        if columns is None:
            columns = {key: key for key in self.get_columns()}
        # Получаем нужные колонки для трансформации.
        needed_columns = {key: columns[key] for key in columns if key in self.get_columns()}
        # Проверяем, что все эти колонки внутри датасета
        reversed_columns = {v: k for k, v in needed_columns.items()}
        self._check_columns(df, reversed_columns)
        # Получаем нужны преобразования перед основным, и запоминаем новые колонки.
        col_to_del = self._make_before_transforms(df, needed_columns)
        # Переименовываем
        df.rename(columns=reversed_columns, inplace=True)
        # Трансформируем
        if not ignore_optimize:
            self.optimize(df)
        self._transform(df)
        # Обратно переименовываем
        df.rename(columns=needed_columns, inplace=True)
        # Удаляем новые колонки
        for name in df.columns:
            if name in col_to_del:
                del df[name]

    def about(self) -> str:
        """
        Return a string that describes the transformation of the columns in the DataFrame.

        Returns:
            str: A string containing the required columns and the resulting columns after the transformation.
        """
        s = f"""Преобразование требует следующих колонок:
            {self.get_columns()}
            В результате получатся такие колонки:
            {self.out_columns}"""
        return s


class MultiTransform:
    """Класс для применения сразу нескольких преобразований."""

    def __init__(self, transforms: list[Transform]) -> None:
        """Init."""
        self.transforms = transforms

    def transform(self, df: pd.DataFrame, name_pattern: Union[dict[str, str], None] = None) -> None:
        """Применяет все преобразования, при этом передавая нужные названия колонок."""
        for transform in self.transforms:
            transform.transform(df, name_pattern)

    def get_columns(self) -> dict[str, str]:
        """Возвращает все колонки, которые потребуется для проведения этого преобразования."""
        columns = {}
        for transform in self.transforms:
            columns.update(transform.get_columns())
        return columns

    def about(self) -> str:
        """Описывает, какие колонки нужны будут для проведения преобразования и какие колонки получатся."""
        s = f"""Преобразование требует следующих колонок:
            {self.get_columns()}
            В результате Получатся такие колонки:
            {self.out_columns}"""
        return s


class CoordinatesTransform(Transform):
    """Преобразует координаты, переводя их из экваториальных в галактические."""

    def __init__(self, b_deg: bool = True, l_deg: bool = True, ra_col: str = 'ra', dec_col: str = 'dec') -> None:
        """Init."""
        self.in_columns = {ra_col: 'Ra degree of object',
                           dec_col: 'Dec degree of object'}
        self.ra_col = ra_col
        self.dec_col = dec_col
        self.b_deg = b_deg
        self.l_deg = l_deg
        self.out_columns = []

        if self.b_deg:
            self.out_columns.append('b_deg')
        if self.l_deg:
            self.out_columns.append('l_deg')

        self.before_transforms = dict()

    def _transform(self, df: pd.DataFrame) -> None:
        """
        Transform the given DataFrame by adding the following columns: ['b_deg', 'l_deg'].

        Note:
            Transform goes inplace. Columns ['b_deg', 'l_deg'] might be added.
        """
        c = SkyCoord(ra=np.array(df[self.ra_col]) * u.degree, dec=np.array(df[self.dec_col]) * u.degree)
        if self.b_deg:
            df['b_deg'] = abs(c.galactic.b.deg)
        if self.l_deg:
            df['l_deg'] = abs(c.galactic.l.deg)

    def optimize(self, df: pd.DataFrame) -> None:
        """Choose whether to generate column or stay it as it."""
        self.b_deg = self.b_deg and 'b_deg' not in df.columns
        self.l_deg = self.l_deg and 'l_deg' not in df.columns


class AstronometryTransform(Transform):
    """
    Класс для генерации основных астрометрических признаков.

    * dictance - дистанция между объектами в угловых секундах

    * r98 - радиус для рентгеновского источника, внутри которого с шансом 98% будет верная его локализация.

    * b_deg - координата b в галактической системе координат.

    * l_deg - координата l в галактической системе координат.

    * distance_in_error - координата b в галактической системе координат.
    """

    def __init__(self, distance: bool = True, r98: bool = True, b_deg: bool = True, l_deg: bool = True,
                 distance_in_error: bool = True) -> None:
        """
        Initialize the class with the specified parameters.

        Args:
            distance: Determines whether to include the 'distance' column in the output.
            r98: Determines whether to include the 'r98' column in the output.
            b_deg: Determines whether to include the 'b_deg' column in the output.
            l_deg: Determines whether to include the 'l_deg' column in the output.
            distance_in_error: Determines whether to include the 'dist_in_err' column in the output.
        """
        self.distance = distance
        self.r98 = r98
        self.b_deg = b_deg
        self.l_deg = l_deg
        self.distance_in_error = distance_in_error

        self.in_columns = {'ra_xray': 'Ra degree of xray source',
                           'dec_xray': 'Dec degree of xray source',
                           'ra_opt': 'Ra degree of optical source',
                           'dec_opt': 'Dec degree of optical source'}

        self.before_transforms = dict()

        self.out_columns = []
        if self.distance:
            self.out_columns.append('distance')
        if self.b_deg:
            self.out_columns.append('b')
        if self.l_deg:
            self.out_columns.append('l_deg')
        if self.distance_in_error:
            self.out_columns.append('dist_in_err')
        if self.r98:
            self.out_columns.append('r98')

    def optimize(self, df: pd.DataFrame) -> None:
        """Choose whether to generate column or stay it as it."""
        self.distance = self.distance and 'distance' not in df.columns
        self.b_deg = self.b_deg and 'b' not in df.columns
        self.l_deg = self.l_deg and 'l_deg' not in df.columns
        self.distance_in_error = self.distance_in_error and 'dist_in_err' not in df.columns
        self.r98 = self.r98 and 'r98' not in df.columns

    def _transform(self, df: pd.DataFrame) -> None:
        """
        Transform the given DataFrame by adding some of the ['distance', 'r98', 'b', 'l', 'distance_in_err'].

        Note:
            Transform goes inplace. 5 new columns might be added: ['distance', 'r98', 'b', 'l', 'distance_in_err'].
        """
        c1 = SkyCoord(ra=np.array(df['ra_xray']) * u.degree, dec=np.array(df['dec_xray']) * u.degree)
        c2 = SkyCoord(ra=np.array(df['ra_opt']) * u.degree, dec=np.array(df['dec_opt']) * u.degree)
        sep = c1.separation(c2)

        if self.distance:
            df['distance'] = sep.arcsecond
        if self.b_deg:
            df['b'] = abs(c1.galactic.b.deg)
        if self.l_deg:
            df['l'] = abs(c1.galactic.l.deg)
        if self.distance_in_error:
            df['dist_in_err'] = sep.arcsecond / df['poserr']
        if self.r98:
            df['r98'] = (-2 * (df['poserr']**2) * np.log(1 - 0.98))**0.5


class NeighboursTransform(Transform):
    """Класс для вычисления количества соседей в радиусе целевого источника. Можно задавать разные радиусы."""

    def __init__(self, to_r98: bool = True, to_distance: tuple[float, ...] = (5., 10., 15.)) -> None:
        """
        Initialize an instance of the class.

        Args:
            to_r98: Determines whether to include the 'pairs_r98' column in the output.
            to_distance: A tuple of distance values used to generate the number of neighbours in this range.
        """
        self.to_r98 = to_r98
        self.to_distance = to_distance

        self.in_columns = {}

        self.before_transforms = {('distance', 'r98'): AstronometryTransform(True, True, False, False, False)}

        self.out_columns = []
        if to_r98:
            self.out_columns.append('pairs_r98')
        for i in to_distance:
            self.out_columns.append(f'pairs_{i}')

    def optimize(self, df: pd.DataFrame) -> None:
        """Choose whether to generate column or stay it as it."""
        self.to_r98 = self.to_r98 and 'to_r98' not in df.columns
        list_ = []
        for i in self.to_distance:
            if f'pairs_{i}' not in df.columns:
                list_.append(i)
        self.to_distance = tuple(list_)

    def _transform(self, df: pd.DataFrame) -> None:
        """
        Transform the given DataFrame by adding the columns about number of neighbours.

        Notes:
            Transform goes inplace. Several columns of number in some range might be added.
        """
        def join_neighbours(data: pd.DataFrame, query: str) -> pd.DataFrame:
            neighbours = data.query(query)['id_xray'].value_counts().reset_index()
            neighbours.columns = ['id_xray', 'neighbours']
            return data.join(neighbours.set_index('id_xray'), on='id_xray', how='left')['neighbours'].fillna(value=0).astype(int)

        if self.to_r98:
            df['pairs_r98'] = join_neighbours(df, 'distance<=r98')
        for dist in self.to_distance:
            df[f'pairs_{dist}'] = join_neighbours(df, f'distance<={dist}')


class AstroFactorTransform(Transform):
    """TODO."""

    def __init__(self, frequency_between: tuple[float, float] = (10, 15)) -> None:
        """
        Initialize an instance of the class.

        Parameters:
            frequency_between: Two radius between which the density of the field of optical sources is calculated
        """
        self.frequency_between = frequency_between

        self.in_columns = {'poserr': "Positional error of xray source"}

        self.before_transforms = {
            ('distance',): AstronometryTransform(True, False, False, False, False),
            (f'pairs_{frequency_between[0]}', f'pairs_{frequency_between[1]}'): NeighboursTransform(False, frequency_between)
        }

        self.out_columns = ['astro_factor']

    def _calc_p_match(self, p_c: float, distance: np.ndarray, sigma: np.ndarray, density: np.ndarray) -> np.ndarray:
        """Calculate probability of matching."""
        exp = p_c * np.exp(-distance**2 / (2 * sigma**2))
        uniform = 2 * np.pi * density * sigma**2
        return exp / (exp + uniform)

    def _transform(self, df: pd.DataFrame) -> None:
        """
        Transform the given DataFrame by calculating the distance factor based on the provided parameters.

        Note:
            Transform goes inplace. Add 'distance_factor' columns.
        """
        area = (self.frequency_between[1] ** 2 - self.frequency_between[0] ** 2) * np.pi
        freq = df[f'pairs_{self.frequency_between[1]}'] - df[f'pairs_{self.frequency_between[0]}']

        df['distance_factor'] = self._calc_p_match(1, df['distance'], df['poserr'], freq / area)


class IgnoreTransform(Transform):
    """Класс для удаления заданных полей данных."""

    def __init__(self, ignored_features: list[str] = []) -> None:
        """
        Initialize a new instance of the class.

        Args:
            ignored_features: A list of features to be ignored.
        """
        self.in_columns = dict()
        self.before_transforms = dict()
        self.out_columns = []

        self.ignored_features = ignored_features

    def _transform(self, df: pd.DataFrame) -> None:
        """
        Drop the specified columns from the given DataFrame.

        Note:
            Transform goes inplace. Drop some columns from dataset.
        """
        df.drop(columns=self.ignored_features, inplace=True)


class FluxesTransform(Transform):
    """Класс для преобразования потоков и их ошибок в астрономические единицы."""

    def __init__(self, fluxes: list[Union[tuple[str, str, str], tuple[str, str]]],
                 column_prefix: str = 'mag', is_error: bool = True, mag_shift: float = 0) -> None:
        """
        Initialize an instance of the class.

        Args:
            fluxes: A list of tuples containing information about fluxes.
                Each tuple should contain the name of the flux, the name of the flux error, and a description.
                You can skip the description if you don't want to use it.
                Example [('flux_1', 'flux_1_err'), ('flux_2', 'flux_2_err', 'Flux in 1')].
            column_prefix: A prefix to be added to the output column names. Defaults to 'mag'.
            is_error: A boolean indicating whether the second column is error. If False, it is inverse variance.
            mag_shift: A shift to be added to the fluxes. Defaults to 0.
        """
        self.mag_shift = mag_shift
        self.fluxes = []
        self.is_error = is_error

        self.in_columns = dict()
        self.out_columns = []

        for obj in fluxes:
            if len(obj) == 2:
                flux, error = obj  # type: ignore
                description = flux
            elif len(obj) == 3:
                flux, error, description = obj  # type: ignore
            else:
                raise ValueError(f'Invalid flux tuple: {obj}')

            self.in_columns[flux] = f'Flux column. Description: {description}'
            self.in_columns[error] = f'Flux error column. Description: {description}'
            self.out_columns.append(f'{column_prefix}_{flux}')
            self.fluxes.append((flux, error))

        self.column_prefix = column_prefix
        self.before_transforms = dict()

    def _asinhmag_dm(self, flux: np.ndarray, flux_err: Union[np.ndarray, None] = None,
                     flux_ivar: Union[np.ndarray, None] = None) -> np.ndarray:
        """
        Calculate asinh mognitude with dm shift.

        Args:
            flux: Flux in [nanomaggies].
            flux_err: Flux error in [nanomaggies]. Default is None.
            flux_ivar: Inverse variance of flux in [1/nanomaggies**2]. Default is None.
        Returns:
            The calculated asinh magnitude with dm shift.
        """
        dm = self.mag_shift
        assert (flux_err is not None) ^ (flux_ivar is not None), 'specify only flux_err or flux_ivar'
        f = flux / 1e9 * np.power(10, 0.4 * dm)
        if flux_ivar is not None:
            b = np.power(flux_ivar, -0.5) / 1e9 * np.power(10, 0.4 * dm)
        else:
            if flux_err is None:
                raise Exception('specify at least on of flux_err or flux_ivar')
            b = flux_err / 1e9 * np.power(10, 0.4 * dm)
            f, b = f.astype(np.float64), b.astype(np.float64)  # otherwise type error like
        # TypeError: loop of ufunc does not support argument 0 of type
        # numpy.float64 which has no callable arcsinh method

        return (np.arcsinh(f / (2 * b)) + np.log(b)) * (-2.5 / np.log(10))

    def _transform(self, df: pd.DataFrame) -> None:
        """
        Transform the given dataframe by applying the asinhmag_dm function to the specified flux and error columns.

        Note:
            Transform goes inplace. Columns of new mag might be added.
        """
        if self.is_error:
            for flux, error in self.fluxes:
                name = f'{self.column_prefix}_{flux}'
                df[name] = self._asinhmag_dm(df[flux], df[error])
                df[name].replace([np.inf, -np.inf], np.nan, inplace=True)
        else:
            for flux, ivar in self.fluxes:
                name = f'{self.column_prefix}_{flux}'
                df[name] = self._asinhmag_dm(df[flux], None, df[ivar])
                df[name].replace([np.inf, -np.inf], np.nan, inplace=True)


class ColorTransform(Transform):
    """Class for calculating the color based on the specified magnitudes. Color is calculated as magnitude_1 - magnitude_2."""

    def __init__(self, magnitudes: Union[list[str], None] = None, fluxes: Union[list[str], None] = None,
                 flux_prefix: str = 'mag') -> None:
        """
        Initialize an instance of the class.

        Args:
            magnitudes: list of magnitudes to calculate the color between it.
            fluxes: name of fluxes. You can specifie it, if you use this transform after FluxesTransform and you don't\
                know the final magnitudes column's name. Write here name of flux column, witch you gave to FluxesTransform.
            flux_prefix: a prefix which you gave to FluxesTransform.
        """
        if magnitudes is None and fluxes is None:
            raise Exception('specify at least one of magnitudes or fluxes')
        elif magnitudes is not None and fluxes is not None:
            self.magnitudes = magnitudes + list(map(lambda x: f'{flux_prefix}_{x}', fluxes))
        elif magnitudes is not None and fluxes is None:
            self.magnitudes = magnitudes
        elif magnitudes is None and fluxes is not None:
            self.magnitudes = list(map(lambda x: f'{flux_prefix}_{x}', fluxes))

        self.in_columns = dict()
        self.before_transforms = dict()
        self.out_columns = []

        for i in self.magnitudes:
            if magnitudes is not None:
                self.in_columns[i] = f'Magnitude: "{i}"'
            else:
                self.in_columns[i] = f'Magnitude from following flux column: "{i}"'

        for i, j in list(itertools.combinations(self.magnitudes, 2)):
            self.out_columns.append(f'color_{i}_{j}')

    def _transform(self, df: pd.DataFrame) -> None:
        """
        Transform the given dataframe by calculating the color based on the specified magnitudes.

        Note:
            Transform goes inplace. Columns of new color might be added.
        """
        for i, j in list(itertools.combinations(self.magnitudes, 2)):
            df[f'color_{i}_{j}'] = df[i].values - df[j].values


class AutoTransform:
    """Класс для автоматического анализа данных и подбора преобразований. Создает MultiTransform с подходящими преобразованиями."""

    def __init__(self, ignore_fluxes: bool = True, max_distance: int = 15, add_color_with_fluxes: bool = True,
                 fluxes_col: Union[list[str], None] = None, fluxes_err_col: Union[list[str], None] = None,
                 mag_col: Union[list[str], None] = None, skip_col: list = []) -> None:
        """
        Initialize an instance of the class.

        Args:
            ignore_fluxes: if True, fluxes will be ignored. After transform they won't be presented.
            max_distance: maximum distance between two points. According to this information the ranges of finding neighbours will be calculated.
            add_color_with_fluxes: if True, color will be calculated with fluxes. After geting all magnitudes, all of them will be used to create color.
            fluxes_col: You can additionaly provide fluxes columns. If not, they will be found if they contain 'flux' of 'Flux' substring. Futhermore, found columns should have error columns.
            fluxes_err_col: If you decide to use fluxes_col, you should provide error columns.
            mag_col: You can additionaly provide magnitudes columns. If not, they will be found if they contain 'mag' substring.
            skip_col: You can skip some columns to analize. They won't be used in any of the transforms.
        """
        self.ignore_fluxes = ignore_fluxes
        self.max_distance = max_distance
        self.add_color_with_fluxes = add_color_with_fluxes
        self.fluxes = fluxes_col
        self.fluxes_err = fluxes_err_col
        self.mag = mag_col

        self.skip_col = skip_col

    def get_transforms(self, df: pd.DataFrame) -> MultiTransform:
        """Analize the dataset and return MultiTransform with suitable transforms."""
        columns = list(filter(lambda x: x not in self.skip_col, df.columns))
        distances = np.array([5, 10, 15, 20, 30])
        distances = distances[distances <= self.max_distance]
        if self.fluxes is None:
            self.fluxes = list(filter(lambda x: ('flux' in x or 'Flux' in x) and 'err' not in x, columns))
        if self.fluxes_err is None:
            self.fluxes_err = list(filter(lambda x: ('flux' in x or 'Flux' in x) and 'err' in x, columns))
        if self.mag is None:
            self.mag = list(filter(lambda x: 'mag' in x, columns))
        if self.add_color_with_fluxes:
            self.mag += list(map(lambda x: 'mag_' + x, self.fluxes))
        self.fluxes.sort()
        self.fluxes_err.sort()
        self.mag.sort()
        transforms = [AstronometryTransform(),
                      NeighboursTransform(to_r98=True, to_distance=tuple(distances)),
                      AstroFactorTransform((self.max_distance, max(self.max_distance * 0.5, self.max_distance - 5))),
                      FluxesTransform(list(zip(self.fluxes, self.fluxes_err)), column_prefix='mag'),
                      ColorTransform(magnitudes=self.mag)]
        if self.ignore_fluxes:
            transforms += [IgnoreTransform(self.fluxes + self.fluxes_err)]
        return MultiTransform(transforms)




from .custom import CustomModelContainer
from .prob_transformer import ProbTransformer
from .prob_transformer import ProbTransformerBase

from .optimization import ParamsOptimizator

from .predict_models import CustomCatboostModel



"""Module for checking additional dependencies, which you may install additionaly."""

from importlib.util import find_spec
import warnings
import os


def local_package_path(s: str) -> str:
    """Получения пути внутри пакета из локального пути. Нужно для работы с файлами конфигурации и сохраненными моделями."""
    return os.path.join(os.path.dirname(__file__), s)

def _module_available(module_path: str) -> bool:
    """
    Check if a path is available in your environment.

    >>> _module_available('os')
    True
    >>> _module_available('bla.bla')
    False
    """
    try:
        return find_spec(module_path) is not None
    except ModuleNotFoundError:
        return False


def is_nway_available() -> bool:
    """Check availability of nwaylib."""
    true_case = (
        _module_available("nwaylib")
    )
    if true_case:
        return True
    else:
        warnings.warn("cosmatch[nway] is not available, to install it, run `poetry install --extras 'nway'`")
        return False



import os

import numpy as np
import pandas as pd
import pickle
import astropy
from astropy.table import Table

from typing import Any
import warnings

from ..settings import local_package_path

def local_package_path_(s: str) -> str:
    """Получения пути внутри пакета из локального пути. Нужно для работы с файлами конфигурации и сохраненными моделями."""
    return os.path.join(os.path.dirname(os.path.dirname(__file__)), s)


def read_fits(path: str) -> pd.DataFrame:
    """Читает содержимое fits-файла и переводит его в DataFrame."""
    dat = Table.read(path, format='fits')
    names = [name for name in dat.colnames if len(dat[name].shape) <= 1]
    df = dat[names].to_pandas()
    return df


def save_fits(df: pd.DataFrame, path: str, params_to_fits_file: dict = {}) -> None:
    """Сохраняет DataFrame в FITS файл."""
    data = Table.from_pandas(df)
    data = astropy.io.fits.table_to_hdu(data)
    for key, value in params_to_fits_file.items():
        data.header[key] = value
    data.writeto(path, overwrite=True)


class Saver:
    """Класс для сохранения и загрузки сущностей из файлов через модуль pickle."""

    @staticmethod
    def save(path: str, key: str, object: Any) -> None:
        """Сохранить объект в словарь внутри файла."""
        if os.path.exists(local_package_path(path)):
            with (open(local_package_path(path), "rb")) as file:
                dict_ = pickle.load(file)
                dict_[key] = object
        else:
            dict_ = {key: object}

        with open(local_package_path(path), 'wb') as file:
            pickle.dump(dict_, file)

    @staticmethod
    def load(path: str, key: str) -> Any:
        """Загрузить объект из словаря внутри файла."""
        if os.path.exists(local_package_path(path)):
            with (open(local_package_path(path), "rb")) as file:
                dict_ = pickle.load(file)
        else:
            raise Exception(f"Fail to load file. No {path} in the system. Maybe you missed to save something?")

        if key not in dict_:
            raise Exception(f'There is no {key} in the file {path}.')

        return dict_[key]

    @staticmethod
    def load_like(path: str, start: str) -> list:
        """Загрузить объекты из словаря внутри файла, начинающиеся с заданного шаблона."""
        if os.path.exists(local_package_path(path)):
            with (open(local_package_path(path), "rb")) as file:
                dict_ = pickle.load(file)
        else:
            raise Exception(f"Fail to load file. No {path} in the system. Maybe you missed to save something?")

        keys = list(filter(lambda x: x.startswith(start), dict_))
        res = [(i[len(start):], dict_[i]) for i in keys]

        return res

    @staticmethod
    def load_with_param(path: str, key_to_param: str, param_val: Any) -> dict:
        """Загрузить объекты из словаря внутри файла, начинающиеся с заданного шаблона."""
        if os.path.exists(local_package_path(path)):
            with (open(local_package_path(path), "rb")) as file:
                dict_ = pickle.load(file)
        else:
            raise Exception(f"Fail to load file. No {path} in the system. Maybe you missed to save something?")

        keys = list(filter(lambda x: dict_[x][key_to_param] == param_val, dict_.keys()))
        res = dict()
        for i in keys:  # TODO
            res[i] = dict_[i]
        return res

    @staticmethod
    def delete(path: str, key: str) -> None:
        """Удалить объект из словаря внутри файла."""
        if os.path.exists(local_package_path(path)):
            with (open(local_package_path(path), "rb")) as file:
                dict_ = pickle.load(file)
        else:
            raise Exception(f"Fail to delete key. No {path} in the system. Maybe you missed to save something?")

        if key not in dict_.keys():
            warnings.warn(f'No key = {key} in file. Deleting skipped.')
            return
        del dict_[key]

        with open(local_package_path(path), 'wb') as file:
            pickle.dump(dict_, file)

    @staticmethod
    def save_config(path: str, object: Any) -> None:
        """Сохраняет конфигурацию в файл."""
        with open(local_package_path(path), 'wb') as file:
            pickle.dump(object, file)

    @staticmethod
    def load_config(path: str) -> Any:
        """Загружает конфигурацию из файла."""
        with open(local_package_path(path), 'rb') as file:
            settings = pickle.load(file)
        return settings


"""Future module for classification of sources to different classes."""

import pandas as pd
import numpy as np
from catboost import CatBoostClassifier, Pool
from sklearn.model_selection import train_test_split

from .utils import correlate


def add_classes(data: pd.DataFrame, ignored: list, path_to_class: str = 'DS/LS/LS_class.pickle') -> pd.DataFrame:
    """Initial version of classification."""
    lis = pd.read_pickle(path_to_class)
    classes = correlate(data, lis, 0.3, ('ra_gaia', 'dec_gaia'), ('ra', 'dec'), add_distance=False)
    classes.drop(columns=['id', 'ra', 'dec'], inplace=True)

    X_train, X_test, y_train, y_test = train_test_split(classes.drop(columns=['spectype']), classes['spectype'], stratify=classes['spectype'], test_size=0.2)

    pool_train = Pool(X_train.drop(columns=ignored), y_train)
    pool_test = Pool(X_test.drop(columns=ignored), y_test)

    model = CatBoostClassifier(iterations=1000, depth=6, learning_rate=0.3,
                               loss_function='MultiClass',
                               eval_metric='TotalF1', early_stopping_rounds=50)
    model.fit(pool_train, eval_set=pool_test, verbose=False)

    predict = model.predict_proba(data.drop(columns=ignored))
    data['p_quasar'] = predict[:, 0]
    data['p_galaxy'] = predict[:, 1]
    data['p_star'] = predict[:, 2]
    return data



"""
Модуль, содержащий метрики для разных задач.

Есть следующие метрики:

* RocAucIdentification

* RocAucNearest

* Completeness

* PrecisionIdentification

* RocAucHostless

* PrecisionHostless
"""

import pandas as pd
import numpy as np
from abc import ABC, abstractmethod


from sklearn.metrics import roc_auc_score, precision_recall_curve


class Metric(ABC):
    """
    Абстрактный класс для всех метрик.

    От класса следует наследовать другие абстрактные классы, характеризующиеся на метриках для различных задач.

    На данный момент есть метрики для следующих задач:

    * Задача отождествления: - Поиск верных пар рентген оптика - метрики строятся по уникальным парам рентген-оптикам в наборе.

    * Задача поиска бездомных источников: - Поиск рентгеновских источников, для которых отстутвует оптический компаньон в каталоге - метрика рассчитывается по уникальным рентгеновским источникам.
    """

    def __str__(self) -> str:
        """STR."""
        return self.__class__.__name__

    def __repr__(self) -> str:
        """REPR."""
        return self.__class__.__name__

    def _get_necessary_columns(self, df: pd.DataFrame, columns: list[str] = ['id_xray', 'P_0', 'P_i', 'mark']) -> pd.DataFrame:
        """Оставляет в каталоге только нужные колонки."""
        for col in columns:
            if col not in df.columns:
                raise KeyError(f'{col} should be in columns')
        return df[columns].copy()

    @abstractmethod
    def calculate(self, df: pd.DataFrame) -> float:
        """Подсчет метрики."""
        pass


class MetricIdentification(Metric, ABC):
    """Абстрактный класс для подсчета метрик для задачи поиска отождествлений."""

    pass


class MetricHostless(Metric, ABC):
    """Абстрактный класс для подсчета метрик для задачи поиска бездомных источников."""

    def add_hostless_mark(self, df: pd.DataFrame) -> None:
        """
        Добавляет метку бездомности рентгеновского источника.

        Если метка равна 1 - у источника нет компаньона в наборе данных, если 0 - компаньон есть.
        """
        hostfull = df.groupby('id_xray').max().query('mark==1').index
        df['mark_hostless'] = 1
        df.loc[df['id_xray'].isin(hostfull), 'mark_hostless'] = 0


class RocAucIdentification(MetricIdentification):
    """
    Класс для расчета метрики ROC_AUC для задачи отождествления для всего множества пар.

    Метрика не очень информативна, так как в каталоге присутсвует множество неверных пар, которых легко классифицировать, как неверные пары.
    """

    def calculate(self, df: pd.DataFrame) -> float:
        """
        Расчет метрики.

        Набор данных должен содержать в себе колонки ['id_xray', 'mark', 'P_i'].
        """
        df = self._get_necessary_columns(df, ['id_xray', 'mark', 'P_i'])
        return roc_auc_score(df['mark'], df['P_i'])


class RocAucNearest(MetricIdentification):
    """
    Класс для расчета метрики ROC_AUC для наиболее вероятных пар для задачи поиска отождествлений.

    При подсчете метрики учитываются только пары из множества верных отождествлений и пары с наибольшей расчитанной вероятностью из множества неверных отождествлений.

    Метрика рекомендуется для рассчета, так как откидывает большенство неверных пар с малой вероятностью - а таких пар много (например, объекты с большой дистанцией).
    """

    def calculate(self, df: pd.DataFrame) -> float:
        """
        Расчет метрики.

        Набор данных должен содержать в себе колонки ['id_xray', 'P_i', 'mark'].
        """
        df = self._get_necessary_columns(df, ['id_xray', 'P_i', 'mark'])
        true_pair = df.query('mark==1')
        false_pair = df.query('mark==0')
        group = false_pair.groupby('id_xray').max().rename(
            columns={'P_i': 'tmp'})['tmp']
        false_pair = false_pair.join(
            group, on='id_xray').query('P_i==tmp').drop(
            columns=['tmp'])
        df = pd.concat([true_pair, false_pair])

        return roc_auc_score(df['mark'], df['P_i'])


class Completeness(MetricIdentification, MetricHostless):
    r"""
    Класс для посчета процента правильно найденных верных отождествлений.

    .. math::
        \text{Completeness} = \frac{\text{Количество верно выбранных пар}}{\text{количество всех верных пар рентген-оптика}}

    Где:

    - Количество верно выбранных пар - Для каждого рентгеновского истчоника оставляется пара с наибольшей вероятностью и подсчитывается количество верных пар среди этого множества.

    - Количество всех верных отождествлений рентген-оптика в каталоге.

    Таким образом мы понимаем, какой процент верных пар рентген-оптика из всего множества был выбран, как наиболее верояная пара для рентгеновского источника.

    Метрика Completeness рассчитывается только для верных пар рентген-оптика и не учитывает бездомные рентгеновские источники.
    """

    def calculate(self, df: pd.DataFrame) -> float:
        """
        Расчет метрики.

        Набор данных должен содержать в себе колонки ['id_xray', 'P_i', 'P_0', 'mark'].
        """
        df = self._get_necessary_columns(df, ['id_xray', 'P_i', 'P_0', 'mark'])
        self.add_hostless_mark(df)
        df = df.join(df.groupby('id_xray').max().rename(columns={'P_i': 'tmp'})['tmp'],
                     on='id_xray').query('P_i==tmp').drop_duplicates()

        return np.mean(df['mark'][df['mark_hostless'] == 0])


class PrecisionIdentification(MetricIdentification):
    """
    Класс для подсчета метрики precision при пороге recall=0.9 для задачи отождествления.

    Значение recall не может быть больше метрики completeness.
    """

    def calculate(self, df: pd.DataFrame, level: float = 0.9) -> float:
        """
        Расчет метрики.

        Набор данных должен содержать в себе колонки ['id_xray', 'P_i', 'P_0', 'mark'].
        """
        df = self._get_necessary_columns(df, ['id_xray', 'P_i', 'P_0', 'mark'])
        completeness = Completeness().calculate(df)

        res = precision_recall_curve(df['mark'], df['P_i'])

        res = list(zip(res[0], res[1] * completeness))
        for i in res[::-1]:
            if i[1] > level:
                return i[0]
        return 0


class RocAucHostless(MetricHostless):
    """
    Класс для расчета метрики ROC_AUC для задачи поиска бездомных источников.

    В данном случае каталог группируется по всем рентгеновским источникам и рассматривается ROC_AUC по полю P_0 и показателю отсутсвие отождествления в каталоге.

    Note:
        Метрика будет работать некоректно, если не было произведено добавления бездомных источников.
    """

    def calculate(self, df: pd.DataFrame) -> float:
        """
        Расчет метрики.

        Набор данных должен содержать в себе колонки ['id_xray', 'P_0', 'mark'].
        """
        df = self._get_necessary_columns(df, ['id_xray', 'P_0', 'mark'])
        self.add_hostless_mark(df)
        df = df[['id_xray', 'mark_hostless', 'P_0']].drop_duplicates()

        return roc_auc_score(df['mark_hostless'], df['P_0'])


class PrecisionHostless(MetricHostless):
    """
    Класс для подсчета метрики precision при пороге recall=0.9 для задачи поиска бездомных источников.

    Note:
        Метрика будет работать некоректно, если не было произведено добавления бездомных источников.
    """

    def calculate(self, df: pd.DataFrame, level: float = 0.9) -> float:
        """
        Расчет метрики.

        Набор данных должен содержать в себе колонки ['id_xray', 'P_i', 'P_0', 'mark'].
        """
        df = self._get_necessary_columns(df, ['id_xray', 'P_i', 'P_0', 'mark'])
        self.add_hostless_mark(df)

        df = df[['id_xray', 'mark_hostless', 'P_0']].drop_duplicates()

        res = precision_recall_curve(df['mark_hostless'], df['P_0'])

        res = list(zip(res[0], res[1]))
        for i in res[::-1]:
            if i[1] > level:
                return i[0]
        return 0



"""__init__."""
import warnings
import os
from .utils import local_package_path
warnings.filterwarnings("ignore", message=".*The 'nopython' keyword.*")
warnings.filterwarnings("ignore", message=".*A value is trying to be set on a copy of a slice from a DataFrame.*")

if not os.path.exists(local_package_path('catalogues')):
    os.mkdir(local_package_path('catalogues'))



"""Модуль, в котором реализован класс-контейнер CustomModelContainer, который будет реализовывать модель для решения задачи отождествления."""

import pandas as pd
import numpy as np
from abc import abstractmethod

from ...metrics import Metric
from .prob_transformer import ProbTransformer

from typing import Union, Tuple


class CustomModelContainer:
    """
    predict_model - predict_proba
    """

    def __init__(self, predict_model, probabilities_transformer=ProbTransformer(),  # type: ignore
                 params_optimizator=None, fit_required=False) -> None:  # type: ignore
        """TODO."""
        self.predict_model = predict_model
        self.probabilities_transformer = probabilities_transformer
        self.params_optimizator = params_optimizator
        self.fit_required = fit_required
        self.fitted = False

    def _drop_id(self, X: pd.DataFrame) -> pd.DataFrame:
        """Удалить из набора данных колонки с id объектов."""
        if 'id_opt' in X.columns and 'id_xray' in X.columns:
            return X.drop(columns=['id_opt', 'id_xray'])
        if 'id_opt' in X.columns:
            return X.drop(columns=['id_opt'])
        if 'id_xray' in X.columns:
            return X.drop(columns=['id_xray'])
        return X

    def _predict_proba(self, X: pd.DataFrame) -> pd.DataFrame:
        pred = self.predict_model.predict_proba(self._drop_id(X))
        if len(pred.shape) == 2:
            return pred[:, 1]
        return pred

    def fit(self, X: pd.DataFrame, y: np.ndarray) -> 'CustomModelContainer':
        if self.fit_required:
            self.predict_model.fit(self._drop_id(X), y)
            self.fitted = True
        else:
            Warning.warn('Model does not require fit. Fitting skipped', UserWarning)
        return self

    def predict_(self, X: pd.DataFrame) -> pd.DataFrame:
        if not self.fitted and self.fit_required:
            raise ValueError('Model is not fitted')
        predict = self._predict_proba(X)
        return predict

    def predict_proba(self, X: pd.DataFrame) -> pd.DataFrame:
        predict = self._predict_proba(X)
        df = X[['id_opt', 'id_xray']].copy()
        return self.probabilities_transformer.transform(df, predict)

    def get_predicted(self, X: pd.DataFrame) -> pd.DataFrame:
        predict = self._predict_proba(X)
        return self.probabilities_transformer.transform(X, predict)

    def validate(self, X: pd.DataFrame, y: np.ndarray, metrics: Union[Metric, list[Metric]]) -> dict[str, float]:
        """Валидация модели."""
        pred = self.predict_proba(X)
        pred['mark'] = y

        if not isinstance(metrics, list):
            res = {metrics.__str__(): metrics.calculate(pred)}
        else:
            res = dict()
            for metric in metrics:
                res[metric.__str__()] = metric.calculate(pred)

        return res



"""Модуль, оеализующий преобразование вероятностей."""

from abc import abstractmethod
import pandas as pd


class ProbTransformerBase:

    @abstractmethod
    def __init__(self) -> None:
        """Init."""
        pass

    @abstractmethod
    def transform(self, df: pd.DataFrame, predict: pd.DataFrame) -> pd.DataFrame:
        """Реализует преобразования предсказаний в таблицу с вероятностями."""
        return NotImplemented


class ProbTransformer(ProbTransformerBase):
    """
    Реализует стандартное преобразование вероятностей.

    Превращает вероятности, полученные на основе анализа каждой пары отдельно в пространственные вероятности, которые учитывают всех соседей.

        * p_i - вероятность отождествления рассчитанная отдельно для каждой пары, без учёта пространственной информации. \
            Базовая вероятность после моделей классификации.

        * P_i - перевычисленная вероятность отождествления, с учетом всех соседей для каждого целевого источника.

        * P_0 - вероятность отсутствия отождествления среди всех пар для каждого целевого источника.
    """

    @staticmethod
    def _add_best_flag(df: pd.DataFrame) -> None:
        """Добавить флаг наиболее вероятного отождествления."""
        df['flag_best'] = 0
        df['max_pi'] = df.groupby('id_xray')['P_i'].transform('max')
        df.loc[df['max_pi'] >= df['P_i'] * 0.5, 'flag_best'] = 2
        df.loc[df['max_pi'] == df['P_i'], 'flag_best'] = 1
        del df['max_pi']

    @staticmethod
    def transform(df: pd.DataFrame, predict: pd.DataFrame) -> pd.DataFrame:
        """
        Возвращает таблицу с новыми колонками вероятностей p_i, P_i и P_0.

        Args:
            df: Таблица с набором данных, которые подавались на вход предсказанию модели.
            mark: Результат предсказания модели.
        Returns:
            Таблица df, но с добавленными колонками p_i, P_i и P_0.
        """
        df['p_i'] = predict
        df['P_0'] = df['p_i'] / (1 - df['p_i'] + 1e-9)
        temp = (1 / (df.groupby("id_xray")['P_0'].sum() + 1))
        df.drop(columns=['P_0'], inplace=True)
        df = df.join(temp, on="id_xray")
        df['P_i'] = df['p_i'] / (1 - df['p_i'] + 1e-9) * df['P_0']
        df['P_i'] = df['P_i'].fillna(1)

        ProbTransformer._add_best_flag(df)
        return df



from .data_loader import Retriever
from .data_loader import PS1, Gaia

from .data_loader import CatalogLoader
from .data_loader import CSC2, DESI

from .data_loader import clear_downloaded_files, download_vizier_catalog


__all__ = [
    'Retriever',
    'PS1',
    'Gaia',

    'CatalogLoader',
    'CSC2',
    'DESI',

    'clear_downloaded_files',
    'download_vizier_catalog',
]



import numpy as np
import pandas as pd

from ..utils import add_postfix_to_main_columns, correlate

def generate_catalog(name: str, size: int = 100, random_state: int = 42,
                     other_columns: list = [], num_flux: int = 0,
                     deviation : float = 1) -> pd.DataFrame:
    """
    Генерирует искусственный каталог. На вход принимает название каталога и его размер и дополнительные колонки.
    
    Args:
        name: Название каталога, которое будет использоваться для обозначения колонок id, ra, dec.
        size: Размер каталога - количество источников, которые будут в нем содержаться.
        random_state: np.random.seed().
        other_columns: Список дополнительных колонок, которые будут заполняться случайными числами. Исключение тут будет \
            колонка 'poserr', которая будет браться из нормального распредления (1.7, 1.4). В случае, если вам нужна колонка \
            'poserr', то вам нужно добавить ее в other_columns.
        num_flux: Количество различных потоков в каталоге. Для каждого потока будет создана колонка 'Flux_{i}' и 'Flux_{i}_err',
            где i - это порядковый номер потока. Значения потока примерно соответсвуют потоку рентгеновских источников.
        deviation : Стандартное отклонение для генерации координат. Чем больше значение, тем выше будет их разброс - меньше кучность.

    Returns:
        Каталог источников с указанными в параметрах колонками. В каталоге также будут указаны основные attrs - ['name', 'id', 'ra', 'dec'].

    Examples:
        >>> from cosmatch.fake import generate_catalog
        >>> opt = generate_catalog(name='opt', size=100)
        >>> opt.columns
        >>> # Index(['id_opt', 'ra_opt', 'dec_opt'], dtype='object')
        >>> opt.shape
        >>> # (100, 3)
        >>> opt.attrs
        >>> # {'name': 'opt', 'id': 'id_opt', 'ra': 'ra_opt', 'dec': 'dec_opt'}
        
        Функция также может создавать различные колонки, в том числе poserr и потоки.

        >>> opt = generate_catalog(name=None, size=1000, num_flux=3, other_columns=['some', 'poserr'])
        >>> opt.columns
        >>> # Index(['id', 'ra', 'dec', 'poserr', 'some', 'Flux_1', 'Flux_1_err',
        >>> #        'Flux_2', 'Flux_2_err', 'Flux_3', 'Flux_3_err'],
        >>> #       dtype='object')
        >>> opt.attrs
        >>> # {'name': '', 'id': 'id', 'ra': 'ra', 'dec': 'dec'}
        
        Можно отдельно указать стандартное отклонение для генерации координат.

        >>> xray = generate_catalog(name='xray', size=1000, num_flux=1, other_columns=['poserr'], deviation=100)
        >>> xray['ra_xray'].std().round(5), opt['ra'].std().round(5)
        >>> # (0.29214, 0.00292)

        Чем меньше отклонение, тем более кучно располагаются объекты - таким образом, про сопоставлении каталогов будет получаться\
            больше пар между каталогами.
        
        >>> from cosmatch.utils import correlate
        >>> opt = generate_catalog(name='opt', size=1000, deviation=1)
        >>> xray = generate_catalog(name='xray', size=1000, deviation=1)
        >>> cor = correlate(opt, xray, 1)
        >>> cor.shape
        >>> # (4614, 6)
        >>> 
        >>> opt = generate_catalog(name='opt', size=1000, deviation=10)
        >>> xray = generate_catalog(name='xray', size=1000, deviation=10)
        >>> cor = correlate(opt, xray, 1)
        >>> cor.shape
        >>> # (1028, 6)

        Заметно, что при увеличении deviation в n раз, количество пар уменьшается в sqrt(n) раз. Происходит это по причине \
            двумерности координат.
    """
    np.random.seed(random_state)
    catalod = pd.DataFrame({'id': np.arange(size),
                           'ra': np.random.random(size) / 100 * deviation  + 25,
                           'dec': np.random.random(size) / 100 * deviation  + 50})
    add_postfix_to_main_columns(catalod, name, add_to_attrs=True)

    if 'poserr' in other_columns:
        catalod['poserr'] = np.random.normal(1.7, 1.4, len(catalod))
    for i in other_columns:
        catalod[i] = np.random.random(size)
    for i in range(num_flux):
        catalod[f'Flux_{i + 1}'] = np.random.random(len(catalod)) * 1e-14
        catalod[f'Flux_{i + 1}_err'] = np.random.random(len(catalod)) * 1e-14

    return catalod


def generate_correlated_catalog(name_frame: str = 'frame', name_target: str = 'target', size_frame: int = 500, size_target: int = 200,
                                random_state: int = 42, num_frame_flux: int = 0, num_target_flux: int = 0, 
                                deviation_frame: int = 1,  deviation_target: int = 1,
                                other_columns: list = [],
                                max_distance: float = 15, add_distance: bool = True,
                                add_mark: bool = False, mark_method: str = 'nearest') -> pd.DataFrame:
    """
    Генерирует объединенный каталог из двух каталогов.

    Args:
        name_frame: Названия frame каталога, который будет сопоставлен с target каталогом. Название будет непосредственно влиять на\
            имена колонок ['id', 'ra', 'dec'] и на attrs.
        name_target: Аналогично name_frame, только для target каталога.
        size_frame: Количество объектов frame каталога.
        size_target: Количество объектов target каталога.
        random_state: np.random.seed()
        num_frame_flux: Количество потоков в frame каталоге.
        num_target_flux: Количество потоков в target каталоге.
        other_columns: Дополнительные колонки каталога. Например, ['some']. В данных колонках будут генерироваться случайные значения.
        max_distance: Максимальное расстояние между объектами в парах.
        add_distance: Добавить расстояние между объектами в парах в выходную таблицу.
        add_mark: Добавить ли метку mark правильности сопоставления каталогов. Метка=1, если считается,\
            что пара frame-target является верным отождествлением
        mark_method: Метод сопоставления, на основе которого ставится метка mark ['nearest', 'random'].\
            'nearest' - ставить метку 1 для каждого target источника в паре с ближайшим оптическим источником.\
            'random' - ставить метку 1 для случайного target источника в паре с оптическим источником.

    Returns:
        Совмещенный каталог данных frame и target. Содержит колонки frame, target ('id', 'ra', 'dec', [flux]...)\
            и опционально ['distance', 'mark'].

    Examples:
        По умолчанию генерируются только основные колонки, такие как ['id', 'ra', 'dec'].

        >>> from cosmatch.fake import generate_correlated_catalog
        >>> data = generate_correlated_catalog()
        >>> data.columns
        >>> # Index(['id_frame', 'ra_frame', 'dec_frame', 'id_target', 'ra_target',
        >>> #        'dec_target', 'poserr', 'distance'],
        >>> #       dtype='object')
        >>> data.id_frame.nunique(), data.id_target.nunique()
        >>> # (500, 200)
        >>> data.attrs
        >>> # {'ids': ['id_frame', 'id_target'],
        >>> #  'coords': ['ra_frame', 'dec_frame', 'ra_target', 'dec_target'],
        >>> #  'name_frame': 'frame', 'id_frame': 'id_frame', 'ra_frame': 'ra_frame', 'dec_frame': 'dec_frame',
        >>> #  'name_target': 'target', 'id_target': 'id_target', 'ra_target': 'ra_target', 'dec_target': 'dec_target'}
        
        Вы можете отдельно выбрать названия подкаталогов, потоки в этох подкаталогах и дополнительные колонки.

        >>> data = generate_correlated_catalog(name_frame='opt', name_target='xray', num_frame_flux=2, num_target_flux=2, other_columns=['some'])
        >>> data.columns
        >>> # Index(['id_opt', 'ra_opt', 'dec_opt', 'Flux_1_opt', 'Flux_1_err_opt',
        >>> #        'Flux_2_opt', 'Flux_2_err_opt', 'id_xray', 'ra_xray', 'dec_xray',
        >>> #        'poserr', 'Flux_1_xray', 'Flux_1_err_xray', 'Flux_2_xray',
        >>> #        'Flux_2_err_xray', 'distance', 'some'], dtype='object')
        >>> data['Flux_1_opt'].mean(), data['Flux_1_opt'].min(), data['Flux_1_opt'].max()
        >>> # (5.131165305927816e-15, 4.9399809344096155e-17, 9.994137257706666e-15)
        
        При генерации совмещенного каталога важно указывать нужную дистанцию между парами (`max_distance`). 
        При том вы можете получить и маркировку верных отождествлений при помощи параметров `add_mark` и `mark_method`.

        >>> data = generate_correlated_catalog(name_frame='opt', name_target='xray', max_distance=5, add_mark=True, mark_method='nearest', add_distance=True)
        >>> data['mark'].value_counts()
        >>> # mark
        >>> # 0    7598
        >>> # 1     200
        >>> # Name: count, dtype: int64
        >>> data.query('mark==1')['id_xray'].nunique()
        >>> # 200
    """
    frame = generate_catalog(name_frame, size=size_frame, random_state=random_state, num_flux=num_frame_flux)
    target = generate_catalog(name_target, size=size_target, random_state=random_state+1, other_columns=['poserr'], num_flux=num_target_flux)
    df = correlate(frame, target, max_distance, add_distance=True, add_attrs=True)

    if add_mark:
        if mark_method == 'nearest':
            df['mark'] = 0
            df.loc[df['distance'] <= df.groupby(df.attrs['id_target']).transform('min').reset_index()['distance'], 'mark'] = 1
        elif mark_method == 'random':
            df['mark'] = 0
            df.loc[df['distance'] == df.groupby(df.attrs['id_target']).transform('first').reset_index()['distance'], 'mark'] = 1

    for i in other_columns:
        df[i] = np.random.random(len(df))
    if not add_distance:
        df.drop('distance', axis=1, inplace=True)
    return df




"""Module for correlation several dataframes to one frame."""

import pandas as pd
import numpy as np

from ..utils import add_postfix_to_main_columns, get_id_ra_dec_col, find_all_id_cols
from ..utils import correlate, keep_nearest_pairs

from typing import List


def connection_table_nearest(df_main: pd.DataFrame, df_sub: pd.DataFrame, max_dist: float = 0.5,
                             keep_distance: bool = True) -> pd.DataFrame:
    """
    Correlate two dataframes and keep nearest pairs to all sources in df_main.

    Args:
        df_main: Main dataframe with sources.
        df_sub: Sub dataframe with sources.
        max_dist: Maximum distance between sources in df_main and df_sub.
        keep_distance: Keep distance between sources in df_main and df_sub in the output.

    Returns:
        Correlated dataframe with columns id_main, id_sub and (optional) distance. id_main is unique.
    """
    id_main, ra_main, dec_main = get_id_ra_dec_col(df_main)
    id_sub, ra_sub, dec_sub = get_id_ra_dec_col(df_sub)
    df_cor = correlate(df_main, df_sub, max_dist, (ra_main, dec_main), (ra_sub, dec_sub), add_distance=True)
    df_cor = keep_nearest_pairs(df_cor, id_main)
    # df_cor = keep_nearest_pairs(df_cor, id_main)
    col = [id_main, id_sub]
    if keep_distance:
        col.append('distance')
    return df_cor[col]


def connection_table_after_correlation(df_cor: pd.DataFrame, query: str = "P_0<0.05 and P_i>0.9 and flag_best==1",
                                       keep_distance: bool = True, make_unique: bool = False) -> pd.DataFrame:
    """
    Keep only best sources after smart-correlation.

    Args:
        df_cor: Correlated dataframe with columns ['P_i', 'P_0', 'flag_best'].
        query: Query for filtering.
        keep_distance: Keep distance between sources in df_main and df_sub in the output.
        make_unique: Make id_main unique. If after smart-correlation there are sources with several pairs with flag_best==1,\
            keep only one pair (it may be the result of duplicates in optical catalog).
    Returns:
        Correlated dataframe with columns id_main, id_sub and (optional) distance.
    """
    df = df_cor.query(query)
    col = find_all_id_cols(df)
    if keep_distance:
        col.append('distance')
    if make_unique:
        df = df.sort_values('P_i', ascending=False).drop_duplicates(subset=['id_opt'], keep='first').reset_index(drop=True)
    return df[col].reset_index(drop=True)


def join_connection_table(df: pd.DataFrame, connection_df: pd.DataFrame, joined_id: str, outer_id: str) -> pd.DataFrame:
    """
    Join dataset of sources into connection table.

    TODO.
    """
    joined_inner = df.join(connection_df.set_index(joined_id), on=joined_id, how='inner')
    joined_outer = df[~df[joined_id].isin(joined_inner[joined_id])]
    joined_outer[outer_id] = -1
    joined = pd.concat([joined_inner, joined_outer])
    return joined


def make_frame(connection_tables: List[pd.DataFrame]) -> pd.DataFrame:
    """
    Create frame table with id_main and all other id in df in list.

    To make the frame, all connection table should contain one same column with id of sources (name of the columns should be the same).
    """
    intersection = set(connection_tables[0].columns)
    for i in connection_tables[1:]:
        intersection = intersection.intersection(i)
    id_main = list(intersection)[0]

    ids = {item for table in connection_tables for item in table[id_main]}

    frame = pd.DataFrame({id_main: list(ids)})
    for table in connection_tables:
        id_cols = find_all_id_cols(table)
        id_sub = id_cols[0] if id_cols[0] != id_main else id_cols[1]
        frame = join_connection_table(frame, table, id_main, id_sub)
    return frame


def join_df_in_frame(frame: pd.DataFrame, dfs: List[pd.DataFrame]) -> pd.DataFrame:
    """
    Join all df in list in frame.

    Frame should contain id columns of all df.

    Args:
        frame: Frame with id columns.
        dfs: List of dataframes. Each df should contain id column, which is presented in frame.
    """
    for df in dfs:
        id_in_df = find_all_id_cols(df)[0]
        frame = pd.merge(frame, df, on=id_in_df, how='left')
    return frame

def mark_object_in_catalog(df_main: pd.DataFrame, catalog: pd.DataFrame, distance: float=0.2, hist=False) -> pd.DataFrame:
    """
    Mark rows with sources, which are represented in the second catalog (near the objects from this catalog).

    Args:
        df_main: Dataframe with sources, where is_in attribute will be added. Attrs: ['id', 'ra', 'dec', 'name']
        catalog: Catalog with sources, near which will be marked. Attrs: ['id', 'ra', 'dec', 'name']
        distance: Distance in degrees to find pairs.
    Returns:
        df_main with new column is_in_{catalog.attrs['name']}. Attrs: ['id', 'ra', 'dec', 'name'] + ['is_in']
    Note:
        df_main will be modified in place. New column is_in_{catalog.attrs['name']} will be added.
    """
    cor = correlate(df_main, catalog, distance, add_distance=True)
    cor.distance.hist(bins=min(100, len(cor))) if hist else None
    df_main['is_in_'+catalog.attrs['name']] = df_main[df_main.attrs['id']].isin(cor[df_main.attrs['id']])

    df_main.attrs['is_in'] = [] if 'is_in' not in df_main.attrs else df_main.attrs['is_in'] + [catalog.attrs['name']]
    df_main.attrs['is_in'] = list(set(df_main.attrs['is_in']))

    return df_main



"""Модуль для составления пайплайна с предсказанием."""

import pandas as pd
import numpy as np

from typing import Union

from ..transforms import Transform, IgnoreTransform
from .models import Model
from .metrics import Metric, RocAucNearest, RocAucHostless
from .data_handler import SampleSelection


class Pipeline:
    """TODO."""

    def __init__(self, model: Model, transforms: list[Transform] = [],
                 metrics: list[Metric] = [RocAucNearest(), RocAucHostless()],
                 ignored_features: list[str] = [],
                 how_to_split: str = 'base', add_hostless: str = 'base',
                 df_to_filter: pd.DataFrame = None, random_state: int = 42) -> None:
        """
        TODO
        """
        self.transforms = transforms
        if ignored_features:
            self.transforms.append(IgnoreTransform(ignored_features))
        self.model = model
        self.metrics = metrics
        self.calc_metric: dict[Metric, float] = dict()
        for i in self.metrics:
            self.calc_metric[i] = np.nan
        self.ignored_features = ignored_features
        self.how_to_split = how_to_split
        self.add_hostless = add_hostless
        self.df_to_filter = df_to_filter
        self.random_state = random_state

        self.fitted_columns = None

    def __str__(self) -> str:
        return NotImplemented

    def get_name_pattern(self) -> dict[str, str]:
        patterns = dict()
        for transform in self.transforms:
            patterns.update(transform.get_columns())
        return patterns

    def transform(self, df: pd.DataFrame, name_pattern: Union[dict[str, str], None] = None) -> pd.DataFrame:
        df = df.copy()
        for transform in self.transforms:
            transform.transform(df, name_pattern)
        return df

    def split_data(self, df: pd.DataFrame) -> tuple[pd.DataFrame, pd.DataFrame, np.ndarray, np.ndarray]:
        if (self.how_to_split == 'improved' or self.add_hostless == 'improved')\
                and self.df_to_filter is None:
            Warning.warn('You should set df_to_filter for improved split or add_hostless.\
                          Base split will be used.', UserWarning)
        return SampleSelection.get_splited_data(df, self.how_to_split, self.add_hostless,
                                                seed=self.random_state, df_to_filter=self.df_to_filter)

    def tune_hyperopt(self, df: pd.DataFrame, iters: int = 50, save: bool = False,
                      name: str = 'tmp', verbose: bool = True) -> 'Pipeline':
        X_train, _, y_train, _ = self.split_data(df)
        self.model.optimize_params(X_train, y_train, iters, verbose)
        if save:
            self.model.save_params(name)
        return self

    def calc_metric_on_set(self, df: pd.DataFrame, query: Union[str, None] = None) -> dict[Metric, float]:
        X_train, X_test, y_train, y_test = self.split_data(df)
        if isinstance(query, str):
            X_test['mark'] = y_test
            X_test.query(query, inplace=True)
            y_test = np.array(X_test['mark'])
            del X_test['mark']
            return self.model.validate(X_test.query(query), y_test, self.metrics)
        else:
            return self.model.validate(X_test, y_test, self.metrics)

    def fit(self, df: pd.DataFrame) -> 'Pipeline':
        X_train, X_test, y_train, y_test = self.split_data(df)
        self.fitted_columns = X_train.columns
        self.model.fit(X_train, y_train)

        self.calc_metric = self.model.validate(X_test, y_test, self.metrics)
        return self

    def about(self) -> str:
        s = f"""
    Pipeline:
        - model: {self.model}
        - transforms: {self.transforms}
        - metrics result on test: {self.calc_metric}
        - ignored_features: {self.ignored_features}
        - how_to_split: {self.how_to_split}
        - add_hostless: {self.add_hostless}

    To predict, in your dataframe should be with following columns:"""
        for key, val in self.get_name_pattern().items():
            s += f'\n\t{key}: {val}'
        s += f"""\n
    You need to rewrite values in name_patterns from last dictionary by your columns name and give it in predict function.
            """
        return s

    def predict(self, df: pd.DataFrame, name_pattern: Union[dict[str, str], None] = None, keep_features: bool = False) -> pd.DataFrame:
        # df = self.transform(df, name_pattern)
        # Это надо оптимизировать
        if self.fitted_columns is None:
            predict = self.model.get_predicted(df)
        else:
            predict = self.model.get_predicted(df[self.fitted_columns])  # type: ignore
        if not keep_features:
            return predict[['id_xray', 'ra_xray', 'dec_xray', 'id_opt', 'ra_opt', 'dec_opt', 'poserr', 'distance',
                            'p_i', 'P_i', 'P_0', 'flag_best']]
        return predict



"""INIT."""

import warnings

from ...settings import is_nway_available

from .base import NearestNeighbour
from .base import Model
from .ml_models import Catboost
from .ml_models import RandomForest


if is_nway_available():
    print("WOW, NWay is available!")
    from .nway_models import NWAY
    from .nway_models import NWAYml
    from .nway_models import NWAYauto



"""Различные полезные функции."""

import numpy as np
import pandas as pd
from astropy.coordinates import SkyCoord
from astropy.units import degree, arcsec
import warnings

from .columns_handler import unite_attrs_from_frame_and_target

from typing import Tuple, Union


def correlate(frame: pd.DataFrame, target: pd.DataFrame, deg: float = 0.001,
              frame_cord: Union[Tuple[str, str], None] = None, target_cord: Union[Tuple[str, str], None] = None,
              lsuffix: str = '_opt', rsuffix: str = '_xray',
              add_attrs: bool = False, add_distance: bool = False, ) -> pd.DataFrame:
    """
    Объединяет два каталога - сопоставляет пары близких объектов из двух каталогов.

    Получает на выход таблицу с парами и всеми признаками из двух входных каталогов в каждой паре.

    Args:
        frame: Первый каталог для объединения.
        target: Второй каталог для объединения.
        deg: Максимальный радиус поиска соседей в секундах.
        frame_cord: Название колонок координат в первом каталоге (re,dec).
        target_cord: Название колонок координат во втором каталоге (re,dec).
        range: Нужно ли добавлять дистанцию между парами
    Returns:
        Объединенный каталог со всеми близкими парами и всеми колонками из двух входных датасетов.
    """
    frame.reset_index(drop=True, inplace=True)
    target.reset_index(drop=True, inplace=True)

    if frame_cord is None:
        frame_cord = [frame.attrs['ra'], frame.attrs['dec']] # type: ignore
    if target_cord is None:
        target_cord = [target.attrs['ra'], target.attrs['dec']] # type: ignore

    c1 = SkyCoord(ra=frame[frame_cord[0]].values * degree, # type: ignore
                  dec=frame[frame_cord[1]].values * degree) # type: ignore
    c2 = SkyCoord(ra=target[target_cord[0]].values * degree, # type: ignore
                  dec=target[target_cord[1]].values * degree) # type: ignore
    idxc1, idxc2, d2d1, d3d1 = c2.search_around_sky(c1, arcsec * deg)
    stack = np.concatenate([idxc1[:, None], idxc2[:, None]], axis=1)
    data = pd.DataFrame(stack, columns=['frame', 'target'])
    data = data.join(frame, on="frame").join(target, on="target", rsuffix=rsuffix, lsuffix=lsuffix)
    if add_distance:
        data['distance'] = d2d1.arcsec[:, None]
    if add_attrs:
        unite_attrs_from_frame_and_target(data, frame, target)
    return data.drop(columns=["frame", "target"])


def add_range(df: pd.DataFrame, first_name: Union[Tuple[str, str], None] = None,
              second_name: Union[Tuple[str, str], None] = None) -> None:
    """
    Добавляет в объединенный каталог пар поле с дистанцией между объектами в паре.

    В каталоге появляется дополнительное поле "distance".

    Args:
        df: Каталог с парами оптических объектов.
        first_name: Название координатных колонов первого объекта в паре (ra,dec).
        second_name: Название координатных колонов второго объекта в паре (ra,dec).

    Note:
        Если first_name и second_name не заданы, ищет их автоматически.
        Находит колонки в каталоге по шаблону 'ra_'+name_1, 'dec_'+name_1 и 'ra_'+name_2, 'dec_'+name_2.

        Функция изменяет каталог - добавляет столбец 'distance'.
    """
    if (first_name is None or second_name is None):
        if 'ra_frame' in df.attrs and 'dec_frame' in df.attrs and 'ra_target' in df.attrs and 'dec_target' in df.attrs:
            first_name = (df.attrs['ra_frame'], df.attrs['dec_frame'])
            second_name = (df.attrs['ra_target'], df.attrs['dec_target'])
        else:
            warnings.warn("TODO.")

            ra_first, ra_second = sorted(list(filter(lambda x: 'ra_' in x, df.columns)))
            dec_first, dec_second = sorted(list(filter(lambda x: 'dec_' in x, df.columns)))
            first_name = (ra_first, dec_first)
            second_name = (ra_second, dec_second)

    c1 = SkyCoord(ra=df[first_name[0]].values * degree, dec=df[first_name[1]].values * degree)
    c2 = SkyCoord(ra=df[second_name[0]].values * degree, dec=df[second_name[1]].values * degree)
    sep = c1.separation(c2)
    df["distance"] = sep.arcsecond


def keep_nearest_pairs(df: pd.DataFrame, object_id: str, distance_col: str = 'distance',
                       first_coord_names: Union[Tuple[str, str], None] = None,
                       second_coord_names: Union[Tuple[str, str], None] = None) -> pd.DataFrame:
    """
    Оставляет пары с объектом (object_id) с наименьшей дистанцией до других объектов.

    Args:
        df: Каталог с парами оптических объектов.
        object_id: Идентификатор объекта.
        distance_col: Название колонки с дистанцией. В случае, если она не находится в каталоге - вычисляется.
            Для этого вызывается функция add_range с параметрами
        first_coord_names: Название координатных колонок первого объекта в паре (ra,dec). В случае None - находится автоматически.
        second_coord_names: Название координатных колонок второго объекта в паре (ra,dec). В случае None - находится автоматически.

    Returns:
        Каталог с объектами с наименьшей дистанцией до других объектов.
    """
    if distance_col not in df.columns:
        add_range(df, first_coord_names, second_coord_names)
        df.rename(columns={'distance': distance_col}, inplace=True)

    min_distance = df.groupby(object_id)[distance_col].transform('min')
    return df[df[distance_col] == min_distance].copy()



"""Набор моделей для решения задачи отождествления и поиска бездомных источников."""
import pandas as pd
import numpy as np
from abc import abstractmethod

from ..metrics import Metric
from .custom import ProbTransformer

from typing import Union, Any


class Model:
    """Абстрактная модель для решения задачи отождествления."""

    @abstractmethod
    def __init__(self) -> None:
        """Абстрактный метод для переопределения в наследуемых классах."""
        pass

    @abstractmethod
    def predict_proba(self, df: pd.DataFrame) -> pd.DataFrame:
        """Предсказание вероятностей."""
        pass

    def __str__(self) -> str:
        """String representation."""
        return self.__class__.__name__  # type: ignore

    def __name__(self) -> str:
        """Name."""
        return 'model'

    def fit(self, X: pd.DataFrame, y: np.ndarray) -> 'Model':
        """Обучение модели."""
        return NotImplemented

    def get_predicted(self, X: pd.DataFrame) -> pd.DataFrame:
        """Сделать предсказания модели с 3 вероятностями: p_i, P_i, P_0 и флагом наиболее вероятного отождествления."""
        X = X.copy()
        mark = self.predict_proba(X)
        return ProbTransformer.transform(X, mark)

    def validate(self, X: pd.DataFrame, y: np.ndarray, metrics: Union[Metric, list[Metric]]) -> dict[Metric, float]:
        """Валидация модели."""
        pred = self.get_predicted(X)
        pred['mark'] = y

        if not isinstance(metrics, list):
            res = {metrics: metrics.calculate(pred)}
        else:
            res = dict()
            for metric in metrics:
                res[metric] = metric.calculate(pred)

        return res


class NearestNeighbour(Model):
    """Модель поиска ближайшего соседа. Строит вероятнсти TODO."""

    def __init__(self, frequency_between: tuple[int, int] = (10, 15)) -> None:
        """Инициализация модели."""
        self.frequency_between = frequency_between

    def _join_neighbours(self, data: pd.DataFrame, query: str) -> pd.DataFrame:
        """Строит датасет с колонкой равной количеству соседей в радиусе."""
        neighbours = data.query(query)['id_xray'].value_counts().reset_index()
        neighbours.columns = ['id_xray', 'neighbours']
        return data.join(neighbours.set_index('id_xray'), on='id_xray', how='left')['neighbours'].fillna(value=0).astype(int)

    def _calc_p_match(self, p_c: float, distance: np.ndarray, sigma: np.ndarray, density: np.ndarray) -> np.ndarray:
        """Вычисляет вероятность отождествления."""
        exp = p_c * np.exp(-distance**2 / (2 * sigma**2))
        uniform = 2 * np.pi * density * sigma**2
        return (exp / (exp + uniform)).fillna(value=0)

    def predict_proba(self, X: pd.DataFrame) -> np.ndarray:
        """
        Выполняет метод predict_proba для внутренней модели.

        Args:
            X: Таблица с парами рентген-оптика.
        Returns:
            Возвращает вероятности принадлежности пары к первому классу.
        """
        area = (self.frequency_between[1] ** 2 - self.frequency_between[0] ** 2) * np.pi
        freq = self._join_neighbours(X, f'distance<={self.frequency_between[1]}') - self._join_neighbours(X, f'distance<={self.frequency_between[0]}')
        res = self._calc_p_match(1, X['distance'], X['poserr'], freq / area)
        return res



"""Модуль с реализованными ML моделями градиентного бустинга и случайного леса."""

import pandas as pd
import numpy as np

import os
import catboost as boost
from sklearn.model_selection import train_test_split
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import StratifiedKFold, cross_val_score
from functools import partial
from abc import abstractmethod
from hyperopt import hp, fmin, tpe, Trials, STATUS_OK

import warnings

from .base import Model
from ...utils import Saver

from typing import Tuple, Union, Any


class ModelML(Model):
    """
    Класс наследования для других моделей.

    В наследуемом классе необходимо реализовать метод :func:`__init__()`, в котором в поле self.model добавляется модель.
    Также необходимо реализовать :func:`__str__()`, который будет возвращать название модели.

    В случае если у присвоенной в :func:`__init__()` self.model есть методы:

    * set_params() - загрузить параметры модели,

    * fit() - обучить модель,

    * predict() - получить предсказание модели,

    * predict_proba() - получить предсказание модели в виде вероятностей,

    Тогда другие методы можно не реализовывать самостоятельно.

    В собственно реализованных классах необходимо не забывать удалять колонки с id космических объектов. Для этого можно
    воспользоваться методом :func:`drop_id()`

    Модели могут выдавать 3 вероятности:

    * p_i - вероятность пары быть верным отождествлением независимо от других пар.

    * P_i - вероятность пары быть верным отождествлением, если её нормировать на вероятности всех пары для этого \
        рентгеновского источника.

    * P_0 - вероятность источника быть бездомным. То есть все найденные пары с этим источником неверные отождествления.

    """

    def __init__(self) -> None:
        """Инициализация модели."""
        self.model = None

    @abstractmethod
    def _params_space(self) -> Union[Tuple[dict, dict], None]:
        """
        Абстрактный метод для переопределения в наследуемых классах.

        Возвращает пространство для перебора гиперпараметров с помощью бибилиотеки hyperopt.

        В случае возвращения None перебор гиперпараметров не определен

        Returns:
            dict: Словарь сетки для поиска гиперпараметров.
            dict: Словарь статических гиперпараметров.
        """
        return None

    def set_params(self, params: dict[str, float]) -> 'Model':
        """Устанавливает параметры модели."""
        self.model.set_params(**params)
        return self

    def load_default_params(self) -> 'Model':
        """Устанавливает стандартные гиперпараметры для модели."""
        self.load_params()
        return self

    def get_params(self) -> dict[str, float]:
        """Возвращает параметры модели."""
        return self.model.get_params()

    def save_params(self, name: Union[str, None] = None) -> None:
        """
        Сохраняет набор гиперпараметров модель по заданному имени.

        При пустом имени модель сохраняется с именем 'tmp'.
        """
        if name is None:
            name = 'tmp'
        name = '__' + self.__str__() + '__' + name
        Saver.save(os.path.join("match", "pretrained", "hyperparams.pickle"), name, self.get_params())

    def get_available_params(self) -> list[tuple[str, Any]]:
        """Возвращает словарь сохраненных наборов гиперпараметров."""
        param_list = Saver.load_like(os.path.join("match", "pretrained", "hyperparams.pickle"), f'__{self.__str__()}__')
        return param_list

    def delete_params(self, name: str) -> None:
        """Удаляет набор гиперпарметров из файла с сохранениями."""
        name = '__' + self.__str__() + '__' + name
        Saver.delete(os.path.join("match", "pretrained", "hyperparams.pickle"), name)

    def load_params(self, name: Union[str, None] = None) -> None:
        """Загрузить набор гиперпараметров модели по заданному имени."""
        if name is None:
            name = 'default'
        name = '__' + self.__str__() + '__' + name
        params = Saver.load(os.path.join("match", "pretrained", "hyperparams.pickle"), name)
        self.set_params(params)

    def optimize_params(self, X: pd.DataFrame, y: np.ndarray, iters: int = 50,
                        verbose: bool = True, other_params: dict[str, Any] = {}, score_function: str = 'roc_auc') -> 'Model':
        """
        Находит лучшие гиперпараметры из пространства поиска для модели.

        Args:
            X: Массив признаков.
            y: Массив меток класса.
            iters: Количество точек для поиска.
            verbose: Надо ли выводить прогресс перебора.
        Returns:
            Оптимизированная модель.
        """
        def objective(params: dict[str, float], model: Model, X: pd.DataFrame, y: np.ndarray) -> dict[str, Any]:
            """Функция для оптимизации."""
            self.set_params(params)
            self.set_params(static_params)
            self.set_params(other_params)
            skf = StratifiedKFold(n_splits=4, shuffle=True, random_state=42)
            score = cross_val_score(estimator=model, X=X, y=y, scoring=score_function, cv=skf, n_jobs=None)
            return {'loss': -score.mean(), 'params': params, 'status': STATUS_OK}

        if self._params_space() is None:
            warnings.warn("This model doesn't have space to optimize params. Optimization skiped", Warning)
            return self

        params_space, static_params = self._params_space()  # type: ignore
        trials = Trials()
        best = fmin(fn=partial(objective, model=self.model, X=X, y=y), space=params_space,
                    algo=tpe.suggest, max_evals=iters, trials=trials,
                    show_progressbar=True, verbose=verbose)
        self.set_params(trials.best_trial['result']["params"])
        self.set_params(static_params)
        return self

    def _drop_id(self, X: pd.DataFrame) -> pd.DataFrame:
        """Удалить из набора данных колонки с id объектов."""
        if 'id_opt' in X.columns and 'id_xray' in X.columns:
            return X.drop(columns=['id_opt', 'id_xray'])
        if 'id_opt' in X.columns:
            return X.drop(columns=['id_opt'])
        if 'id_xray' in X.columns:
            return X.drop(columns=['id_xray'])
        return X

    def fit(self, X: pd.DataFrame, y: np.ndarray) -> 'Model':
        """
        Обучить модель на помеченном наборе данных.

        Args:
            X: Таблица пар рентген-оптика с признаками и id источников.
            y: Массив меток классов.
        """
        self.model.fit(self._drop_id(X), y)
        return self

    def predict_proba(self, X: pd.DataFrame) -> np.ndarray:
        """
        Выполняет метод predict_proba для внутренней модели.

        Args:
            X: Таблица с парами рентген-оптика.
        Returns:
            Возвращает вероятности принадлежности пары к первому классу.
        """
        return self.model.predict_proba(self._drop_id(X))[:, 1]


class Catboost(ModelML):
    """Градиентный бустинг."""

    def __init__(self, gpu: bool = False) -> None:
        """Инициализация модели."""
        self.gpu = gpu
        self.model = boost.CatBoostClassifier()
        self.load_default_params()

    def set_params(self, params: dict[str, float]) -> 'Catboost':
        """Устанавливает параметры модели."""
        params_ = {'iterations': 250}
        self.model.set_params(**params)
        self.model.set_params(**params_)
        return self

    def load_default_params(self) -> 'Catboost':
        """Устанавливает стандартные гиперпараметры для модели."""
        if not self.gpu:
            self.load_params('default_cpu')  # REDO
        else:
            self.load_params()
        return self

    def fit(self, X: pd.DataFrame, y: np.ndarray) -> 'Catboost':
        """Обучение модели."""
        X_train, X_test, y_train, y_test = train_test_split(self._drop_id(X),
                                                            y, test_size=0.2, random_state=42)
        params = self.model.get_params()
        self.model = boost.CatBoostClassifier().set_params(**params)
        self.model.set_params(**{'use_best_model': True, 'eval_metric': 'F1'})
        self.model.fit(X_train, y_train, eval_set=(X_test, y_test),
                       use_best_model=True, early_stopping_rounds=50, logging_level='Silent')
        return self

    def _params_space(self) -> Tuple[dict[str, Any], dict[str, Any]]:
        """
        Возвращает пространство для перебора гиперпараметров с помощью бибилиотеки hyperopt.

        Returns:
            dict: Словарь сетки для поиска гиперпараметров.
            dict: Словарь статических гиперпараметров.
        """
        space = {'random_strength': hp.randint('random_strength', 1, 20),
                 'depth': hp.randint('depth', 1, 12),
                 'learning_rate': hp.uniform('learning_rate', np.exp(-7), 1),
                 'l2_leaf_reg': hp.uniform('l2_leaf_reg', 1, 10),
                 'bagging_temperature': hp.uniform('bagging_temperature', 0, 1),
                 'score_function': hp.choice(label='score_function', options=['L2', 'NewtonL2', 'NewtonCosine', 'Cosine']),
                 'leaf_estimation_backtracking': hp.choice(label='leaf_estimation_backtracking',
                                                           options=['No', 'AnyImprovement', 'Armijo']),
                 'leaf_estimation_iterations': hp.randint('leaf_estimation_iterations', 1, 20)}
        static = {
            'loss_function': 'CrossEntropy',
            'task_type': 'GPU',
            'devices': '0:1',
            'iterations': 100,
            'use_best_model': False,
            'logging_level': 'Silent'}

        if not self.gpu:
            static['task_type'] = 'CPU'
            space['leaf_estimation_backtracking'] = hp.choice(label='leaf_estimation_backtracking',
                                                              options=['No', 'AnyImprovement'])
            space['score_function'] = hp.choice(label='score_function', options=['L2', 'Cosine'])
        return space, static


class RandomForest(ModelML):
    """Случайный лес."""

    def __init__(self) -> None:
        """Инициализация модели."""
        self.model = RandomForestClassifier()
        self.load_default_params()

    def _params_space(self) -> Tuple[dict[str, Any], dict[str, Any]]:
        """
        Возвращает пространство для перебора гиперпараметров с помощью бибилиотеки hyperopt.

        Returns:
            dict: Словарь сетки для поиска гиперпараметров
            dict: Словарь статических гиперпараметров
        """
        space = {'max_features': hp.choice(label='max_features', options=['log2', 'sqrt']),
                 'max_depth': hp.randint('max_depth', 10, 20),
                 'min_samples_split': hp.randint('min_samples_split', 1, 20),
                 'min_samples_leaf': hp.randint('min_samples_leaf', 1, 20),
                 'bootstrap': hp.choice(label='bootstrap', options=[True, False]),
                 'criterion': hp.choice(label='criterion', options=['gini', 'entropy', 'log_loss'])}
        static = {'n_jobs': -1,
                  'n_estimators': 1000}
        return space, static



"""Modul for loading data from Vizier."""

from astroquery.vizier import Vizier
import astropy.coordinates as coord
import astropy.units as u

import pandas as pd
import numpy as np
import os

import requests  # type: ignore
import urllib
import json

from abc import ABC, abstractmethod

from ..utils import local_package_path

from typing import Union


class Retriever(ABC):
    """Abstract class for finding near pairs of object from Vizier.

    To implement this class, you need to define the following methods: __init__

    In this method you should:
        - set the path to the catalog from vizier
        - set the columns to load from catalog
        - set the description of columns (optional)
    """

    def __init__(self) -> None:
        """Initialize the class. You should implement this method in your class."""
        self.path: str  # name of catalog in Vizier
        self.columns: Union[dict[str, str], list[str]]  # list of columns in catalog or dict of {column_in_catalog: more_clear_name}
        self.about_columns: Union[dict[str, str], None] = None  # description of columns

    def load_data(self, sourses: pd.DataFrame, columns: list[str] = ['*'], radius_arcsec: float = 10,
                  verbose: bool = False, window: int = 100) -> pd.DataFrame:
        """
        Load data from Vizier.

        Args:
            sourses: catalog of objects with coordinates [ra, dec]
            columns: list of columns to load from catalog. To get available columns, use available_columns().\
                  To load all columns, use ['*']
            radius_arcsec: radius of search in arcsec.
            verbose: if True, print progress.
            window: number of sources to load at once.
        Returns:
            Dataframe with loaded data.
        Note:
            This method generate temporary files for loading data. It also save data.
        """
        if columns == ['*']:
            columns = self.available_columns()
        columns = self._rename_columns(columns)

        if not os.path.exists('tmp_load_data'):
            os.mkdir('tmp_load_data')
        for num in range(len(sourses) // window + 1):
            coords = coord.SkyCoord(ra=list(sourses.ra.values[num * window:(num + 1) * window]),
                                    dec=list(sourses.dec.values[num * window:(num + 1) * window]), unit=(u.deg, u.deg))
            radius = radius_arcsec * u.arcsec

            if verbose:
                print(f"{num} start...", end=' ')
            panstarrs = Vizier(columns=columns, catalog=self.path).query_region(coords, radius=radius)
            if verbose:
                print(f"get...", end=' ')

            for i, result in enumerate(panstarrs):
                neighbors = result[columns]
                neighbors = pd.DataFrame(np.array(neighbors), columns=columns)
                neighbors.to_csv(f'tmp_load_data/{num}.csv', index=False)

            if verbose:
                print(f"{num*window} object finished!", end='\n')

        print('Data collected successfully. Start to merge.')

        df = pd.concat([pd.read_csv(f'tmp_load_data/{num}.csv') for num in range(len(sourses) // window + 1)])

        df.columns = self._unrename_columns(df.columns)
        df.to_csv(f'merged_{len(df)}.csv', index=False)
        print(f'Data merged successfully. It saved as merged_{len(df)}.csv')

        return df

    def available_columns(self) -> list[str]:
        """Get available columns to load."""
        if isinstance(self.columns, dict):
            return list(self.columns.values())
        else:
            return self.columns

    def _rename_columns(self, columns: list[str]) -> list[str]:
        """Rename columns to vizier columns."""
        if isinstance(self.columns, dict):
            dict_ = {val: key for key, val in self.columns.items()}
            for col in columns:
                if col not in dict_:
                    raise KeyError(f'{col} should be in columns. Available columns: {self.available_columns()}')
            return [dict_[col] for col in columns]
        else:
            for col in columns:
                if col not in self.columns:
                    raise KeyError(f'{col} should be in columns. Available columns: {self.available_columns()}')
            return columns

    def _unrename_columns(self, columns: list[str]) -> list[str]:
        """Reverse method for renaming vizier columns to clear columns."""
        if isinstance(self.columns, dict):
            return [self.columns[col] for col in columns]
        else:
            return columns

    def about(self) -> str:
        """Get description of columns. If you haven't implemented self.about_columns, it will return 'About was not implemented yet."""
        if self.about_columns is None:
            return "About was not implemented yet."
        return str(self.about_columns)


class PS1(Retriever):
    """Class for loading PanSTARRS DR1."""

    def __init__(self) -> None:
        """Init."""
        self.path = 'II/349/ps1'
        self.columns = {"objID": 'id', 'RAJ2000': 'ra', 'DEJ2000': 'dec',
                        'rmag': 'rPsfMag', 'e_rmag': 'rPsfMagErr', 'rKmag': 'rKronMag', 'e_rKmag': 'rKronMagErr',
                        'ymag': 'yPsfMag', 'e_ymag': 'yPsfMagErr', 'yKmag': 'yKronMag', 'e_yKmag': 'yKronMagErr',
                        'zmag': 'zPsfMag', 'e_zmag': 'zPsfMagErr', 'zKmag': 'zKronMag', 'e_zKmag': 'zKronMagErr',
                        'imag': 'iPsfMag', 'e_imag': 'iPsfMagErr', 'iKmag': 'iKronMag', 'e_iKmag': 'iKronMagErr',
                        'gmag': 'gPsfMag', 'e_gmag': 'gPsfMagErr', 'gKmag': 'gKronMag', 'e_gKmag': 'gKronMagErr', }
        self.about_columns = {'id': "Id of the object",
                              'RAJ2000': "Right ascension",
                              'DEJ2000': "Declination",
                              'gPsfMag': "Mean PSF magnitude from g filter",
                              'gPsfMagErr': "Uncertainty in g magnitude from g filter",
                              'gKronMag': "Mean Kron magnitude from g filter detections",
                              'gKronMagErr': "Uncertainty in Kron magnitude from g filter",
                              'rPsfMag': "Mean PSF magnitude from r filter",
                              'rPsfMagErr': "Uncertainty in r magnitude from r filter",
                              'rKronMag': "Mean Kron magnitude from r filter",
                              'rKronMagErr': "Uncertainty in Kron magnitude from r filter",
                              'iPsfMag': "Mean PSF magnitude from i filter",
                              'iPsfMagErr': "Uncertainty in i magnitude from i filter",
                              'iKronMag': "Mean Kron magnitude from i filter",
                              'iKronMagErr': "Uncertainty in Kron magnitude from i filter",
                              'zPsfMag': "Mean PSF magnitude from z filter",
                              'zPsfMagErr': "Uncertainty in i magnitude from z filter",
                              'zKmag': "Mean Kron magnitude from z filter",
                              'zKronMagErr': "Uncertainty in Kron magnitude from z filter",
                              'yPsfMag': "Mean PSF magnitude from y filter",
                              'yPsfMagErr': "Uncertainty in y magnitude from y filter",
                              'yKronMag': "Mean Kron magnitude from y filter",
                              'yKronMagErr': "Uncertainty in Kron magnitude from y filter"}


class Gaia(Retriever):
    """Class for loading GAIA DR3."""

    def __init__(self) -> None:
        """Init."""
        self.path = 'I/355/gaiadr3'
        self.columns = {'Source': 'id', 'RA_ICRS': 'ra', 'DE_ICRS': 'dec',
                        'Plx': 'paralax', 'e_Plx': 'paralaxErr',
                        'Gmag': 'gMag', 'e_Gmag': 'gMagErr',
                        'BPmag': 'bpMag', 'e_BPmag': 'bpMagErr',
                        'RPmag': 'rpMag', 'e_RPmag': 'rpMagErr', }


class CatalogLoader:
    """Class for loading catalogs from yandex disk."""

    def __init__(self, file_name: str) -> None:
        """Init."""
        self.folder_link = 'https://disk.yandex.ru/d/dxNUCwXCW5_mHg'
        self.path_to_save = local_package_path(f'downloaded_catalogues/{file_name.replace("/","_")}')
        self.file_name = file_name
        self.columns: Union[dict, None] = None
        self.str_about: Union[str, None] = None

    def _load_file(self) -> pd.DataFrame:
        """Download file."""
        url = 'https://cloud-api.yandex.net/v1/disk/public/resources/download' + '?public_key='\
              + urllib.parse.quote(self.folder_link) + '&path=/' + urllib.parse.quote(self.file_name)

        r = requests.get(url)  # запрос ссылки на скачивание
        h = json.loads(r.text)['href']  # 'парсинг' ссылки на скачивание
        return pd.read_pickle(h)

    def remove(self) -> None:
        """Delete file from local environment."""
        if os.path.exists(self.path_to_save):
            print(f'File {self.path_to_save} already exists. Removing...')
            os.remove(self.path_to_save)

    def load(self, force: bool = False) -> pd.DataFrame:
        """Download file if in local environment it does not exist."""
        if os.path.exists(self.path_to_save) and not force:
            print(f'File {self.path_to_save} already exists. Loading...')
            return pd.read_pickle(self.path_to_save)

        print(f'File {self.path_to_save} does not exist. Downloading...')
        df = self._load_file()
        df.to_pickle(self.path_to_save)

        return df

    def about(self) -> str:
        """Get description of columns. If you haven't implemented self.about_columns, it will return 'About was not implemented yet."""
        if self.str_about is None:
            return "About was not implemented yet."
        return self.str_about


class CSC2(CatalogLoader):
    """Class for download CSC2 as support catalogue to teach model of correlation."""

    def __init__(self) -> None:
        """Init."""
        super().__init__('CSC2.pkl')


class DESI(CatalogLoader):
    """Class for download DESI LIS, in which DESI classes are stored. Needed for teach model of classification."""

    def __init__(self) -> None:
        """Init."""
        super().__init__('DESI_classes.pkl')


# Переодически бывают ошибки в скачивании данных. Надо пробовать несколько раз


def clear_downloaded_files() -> None:
    """Delete all files in local folder."""
    for file in os.listdir(local_package_path('downloaded_catalogues')):
        if file.endswith('.pkl'):
            os.remove(local_package_path(f'downloaded_catalogues/{file}'))


def download_vizier_catalog(catalog_keyword: str, columns: list[str] = ['**']) -> pd.DataFrame:
    """Download whole catalog from Vizier with specified columns."""
    data = Vizier(columns=columns,
                  catalog=catalog_keyword,
                  row_limit=-1).get_catalogs(catalog=catalog_keyword)[0]
    data = data.to_pandas()
    return data



from .columns_handler import add_postfix_to_main_columns
from .columns_handler import check_attrs_structure
from .columns_handler import get_attrs_columns
from .columns_handler import find_all_id_cols
from .columns_handler import get_id_ra_dec_col
from .columns_handler import unite_attrs_from_frame_and_target

from .io import local_package_path
from .io import read_fits, save_fits
from .io import Saver

from .astrometry import correlate
from .astrometry import add_range
from .astrometry import keep_nearest_pairs


__all__ = [
    'local_package_path',
    'read_fits',
    'save_fits',
    'Saver',

    'correlate',
    'add_range',
    'keep_nearest_pairs',
    
    'add_postfix_to_main_columns',
    'check_attrs_structure',
    'get_attrs_columns',
    'find_all_id_cols',
    'get_id_ra_dec_col',
    'unite_attrs_from_frame_and_target'
]



import pandas as pd

from typing import List, Union


def add_postfix_to_main_columns(df: pd.DataFrame, postfix: Union[str, None], add_to_attrs: bool = True) -> None:
    """
    Добавляет к имени колонок id, ra, dec постфикс postfix. Также позволяет добавить атрибуты ['name', 'id', 'ra', 'dec'].
    
    Args:
        df: DataFrame, к которому добавляются постфиксы. Columns = ['id', 'ra', 'dec']
        postfix: Постфикс для добавляемых колонок. Может быть None,\
            тогда колонки не переименовываются, но атрибуты могут добавиться
        add_to_attrs: Если True, то добавляются атрибуты ['name', 'id', 'ra', 'dec'].

    Examples:
        Важно обратить внимание, что для работы функции в наборе данных должны быть строго колонки ['id', 'ra', 'dec'].

        >>> import pandas as pd
        >>> from cosmatch.utils import add_postfix_to_main_columns
        >>>
        >>> frame = pd.DataFrame({'id': [1, 2, 3, 4, 5], 
        ...                       'ra': [1, 2, 3, 4, 5], 
        ...                       'dec': [1, 2, 3, 4, 5]})
        >>> add_postfix_to_main_columns(frame, postfix='frame', add_to_attrs=True)
        >>> frame.columns
        Index(['id_frame', 'ra_frame', 'dec_frame'], dtype='object')
        >>> frame.attrs
        {'name': 'frame', 'id': 'id_frame', 'ra': 'ra_frame', 'dec': 'dec_frame'}

        В случае передачи постфикса None, то атрибуты добавляются, но колонки не переименовываются.

        >>> frame = pd.DataFrame({'id': [1, 2, 3, 4, 5],
        ...                       'ra': [1, 2, 3, 4, 5],
        ...                       'dec': [1, 2, 3, 4, 5]})
        >>> add_postfix_to_main_columns(frame, postfix=None, add_to_attrs=True)
        >>> frame.columns
        Index(['id', 'ra', 'dec'], dtype='object')
        >>> frame.attrs
        {'name': '', 'id': 'id', 'ra': 'ra', 'dec': 'dec'} 
    """
    if postfix is not None:
        df.rename(columns={'id': f'id_{postfix}', 'ra': f'ra_{postfix}',
                        'dec': f'dec_{postfix}'}, inplace=True)
        if add_to_attrs:
            df.attrs['name'] = postfix
            df.attrs['id'] = f'id_{postfix}'
            df.attrs['ra'] = f'ra_{postfix}'
            df.attrs['dec'] = f'dec_{postfix}'
    else:
        if add_to_attrs:
            df.attrs['name'] = ''
            df.attrs['id'] = 'id'
            df.attrs['ra'] = 'ra'
            df.attrs['dec'] = 'dec'

def unite_attrs_from_frame_and_target(df: pd.DataFrame, frame: pd.DataFrame, target: pd.DataFrame) -> None:
    check_attrs_structure(frame, ['name', 'id', 'ra', 'dec'])
    check_attrs_structure(target, ['name', 'id', 'ra', 'dec'])
    df.attrs['ids'] = [frame.attrs['id'], target.attrs['id']]
    df.attrs['coords'] = [frame.attrs['ra'], frame.attrs['dec'], target.attrs['ra'], target.attrs['dec']]

    df.attrs['name_frame'] = frame.attrs['name']
    df.attrs['id_frame'] = frame.attrs['id']
    df.attrs['ra_frame'] = frame.attrs['ra']
    df.attrs['dec_frame'] = frame.attrs['dec']

    df.attrs['name_target'] = target.attrs['name']
    df.attrs['id_target'] = target.attrs['id']
    df.attrs['ra_target'] = target.attrs['ra']
    df.attrs['dec_target'] = target.attrs['dec']

def check_attrs_structure(df: pd.DataFrame, keys: list[str]=['id', 'ra', 'dec']) -> None:
    """Проверяет структуру атрибутов в df."""
    for key in keys:
        if key not in df.attrs:
            raise ValueError(f'{key} not in attrs. To use correlation module you should add {key} in attrs.')
        
def get_attrs_columns(df: pd.DataFrame, keys: list[str]=['id', 'ra', 'dec']) -> list[str]:
    """Возвращает список атрибутов в df."""
    return [df.attrs[key] for key in keys]


# Legacy
def find_all_id_cols(df: pd.DataFrame) -> List[str]:
    """Возвращает все колонки, которые начинаются на id."""
    return list(filter(lambda x: 'id_' in x, df.columns))

# Legacy
def get_id_ra_dec_col(df: pd.DataFrame) -> List[str]:
    """Возвращает колонки id, ra, dec из каталога. В нем не можно быть нескольких этих колонок."""
    filtered = list(filter(lambda x: 'id_' in x, df.columns))
    if len(filtered) == 0:
        raise Exception('Id column not found')
    if len(filtered) > 1:
        raise Exception('In catalog more than one id column')
    postfix = filtered[0].replace('id_', '')
    ans = [f'id_{postfix}', f'ra_{postfix}', f'dec_{postfix}']
    for i in ans:
        if i not in df.columns:
            raise Exception(f'{i} not found')
    return [f'id_{postfix}', f'ra_{postfix}', f'dec_{postfix}']


"""Модуль направленный на подготовку данных для решения задачи отождествления."""

import numpy as np
import pandas as pd
import warnings

from ..utils import correlate, add_range
from ..utils import check_attrs_structure, get_attrs_columns, unite_attrs_from_frame_and_target

from typing import Tuple, Union


class OpticalNeighbours:
    """
    Класс для получения позитивного (компаньонов рентгеновских источников) и негативного \
    (не компаньонов рентгеновских источников) класса оптических объектов из тренеровочного оптического каталога. \
    Для отбора используется вспомогательный рентгеновский каталог.

    Отбор производится по следующему алгоритму:

    * Находятся пары рентген-оптика для оптических объектов обоих классов так, что дистанция в парах менее \
        заданного порога угловых секунды.

    * Для каждого рентгеновского источника в парах вычисляется радиус, внутри которого с заданным шансом \
        находящийся там оптический объект не будет компаньоном.

    * Оптические объекты положительного класса отбираются, как объекты в парах с дистанцией меньше вычисленного \
        ранее радиуса. Дополнительно убираются неоднозначности.

    * Оптические объекты отрицательного класса отбираются, как объекты, для которых расстояние до любого \
        рентгеновского соседа больше вычисленного радиуса.

    """

    @staticmethod
    def _add_clear_distance(df: pd.DataFrame, name: str = "id_support", max_distance: float = 15, prob: float = 0.05) -> None:
        """
        Добавляет колонку в DataFrame, в которой содержится дистанция для рентгеновского источника, \
        в котрой с заданной вероятностью будут неверные компаньоны.

        Args:
            df: Таблица с парами frame-target.
            name: Название колонки, в которой содержится id target источника.
            max_distance: Максимальная дистанция в парах в каталоге.float
            prob: Вероятность ошибки в отборе надежных пар.

        Note:
            Функция изменяет каталог - добавляет столбец 'clear_distance'.
        """
        freq = df.groupby(name)[name].transform('count')
        area = np.pi * max_distance ** 2
        df['clear_distance'] = OpticalNeighbours._calc_radius(freq / area, prob)

    @staticmethod
    def _calc_radius(density: np.ndarray, probability: float = 0.05) -> np.ndarray:
        """
        Вычисляет радиус надежного окружения target источника, в котором с заданным шансом будет \
        содержаться неверный frame компаньон. Вычисление происходит исходя из плотности поля \
        target объектов вокруг target источника.

        Args:
            density: Значение плотности поля вокруг target источников.
            probability: Вероятность, что frame объект внутри радиуса будет неверным компаньоном \
                для target источника.
        Returns:
            Значение надежного радиуса.
        """
        return (-np.log(1 - probability) / density / np.pi)**0.5

    @staticmethod
    def _delete_mult_obj(df: pd.DataFrame, name: str = "id_support") -> None:
        """
        Удалить ретгеновские/оптические объекты, которые встречаются в таблице более одного раза.

        Args:
            df: Таблица с парами frame-target.
            name: Название колонки, в которой содержится id нужного объекта.

        Note:
            Функция изменяет каталог - удаляет некоторые строки.
        """
        id_column = df[name].copy()
        id_counts = id_column.value_counts()
        good = id_counts[id_counts == 1].index
        df.query(f'{name} in @good', inplace=True)

    def _get_positive(df: pd.DataFrame) -> pd.DataFrame:
        """
        Получает оптические объекты позитивного класса.

        Выходной каталог содержит в себе колонки для оптического тела и координаты рентгеновского источника из вспомогательного каталога.

        Args:
            df: Каталог пар рентген-оптика. Columns: id, ra, dec, id_support, ra_support, dec_support, clear_distance, distance.

        Returns:
            Каталог пар рентген-оптика. Columns: id, ra, dec, ra_support, dec_support.
        """
        positive = df.query("distance<clear_distance").copy()
        OpticalNeighbours._delete_mult_obj(positive, "id_support")
        OpticalNeighbours._delete_mult_obj(positive, "id")
        positive.drop(columns=['distance', 'id_support', 'clear_distance'], inplace=True)
        positive.reset_index(drop=True, inplace=True)
        return positive

    def _get_negative(df: pd.DataFrame) -> pd.DataFrame:
        """
        Получает оптические объекты негативного класса.

        Выходной каталог содержит в себе только колонки для оптического тела.

        Args:
            df: Таблица с парами frame-target. Columns: id, ra, dec, id_support, ra_support, dec_support, clear_distance, distance.

        Returns:
            Таблица с парами frame-target. Columns: id, ra, dec.

        Note:
            Меняет входной каталог - удаляет строки и столбцы.
        """
        potential_positive = df.query('distance<clear_distance')["id"].unique()
        df.query("distance>clear_distance", inplace=True)
        df.query('id not in @potential_positive', inplace=True)
        df.drop(columns=['distance', 'id_support', 'clear_distance', 'ra_support', 'dec_support'], inplace=True)
        df.drop_duplicates(inplace=True)
        return df

    @staticmethod
    def get_two_class_opt(frame: pd.DataFrame, support: pd.DataFrame,
                          max_distance: float = 15, prob: float = 0.05) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Получение позитивной и негативной выборки frame объектов при помощи позиций target источников из вспомогательного каталога.

        Args:
            frame: Каталог с frame объектами. Columns: id, ra, dec
            support: Каталог с target_support объектами. Columns: id, ra, dec
            max_distance: Максимальное расстояние между frame и target_support объектами.
            prob: Вероятность, что frame объект внутри радиуса будет неверным компаньоном для target источника.

        Returns:
            - Каталог с frame объектами - компаньонами, при том добавит координаты рентгеновского источника,\
              которые нужны для более качественного дальнейшего отбора.
            - Каталог с frame объектами - не компаньонами.

        Examples:
            >>> import numpy as np
            >>> import pandas as pd
            >>> from cosmatch.match.data_handler import OpticalNeighbours
            >>> from cosmatch.fake import generate_catalog
            >>>
            >>> frame_size = 10000; target_size = 500
            >>> frame = generate_catalog(name=None, size=frame_size, random_state=43)
            >>> target_support = generate_catalog(name=None, size=target_size, random_state=44)
            >>> positive, negative = OpticalNeighbours.get_two_class_opt(frame, target_support, 
            ...                                                          max_distance=15, prob=0.05)
            >>> positive.columns, negative.columns
            (Index(['id', 'ra', 'dec', 'ra_support', 'dec_support'], dtype='object'),
            Index(['id', 'ra', 'dec'], dtype='object'))
            >>> positive.shape, negative.shape
            ((41, 5), (956, 3))

            Объекты frame  попадают только в один из наборов данных.

            >>> positive['id'].isin(negative['id']).sum()
            0

            При увеличении prob размер позитивной выборки также увеличивается, при том в нем падает надежность - больше шанс отобрать верную пару

            >>> positive, negative = OpticalNeighbours.get_two_class_opt(frame, target_support, 
            ...                                                          max_distance=15, prob=0.10)
            >>> positive.shape, negative.shape
            ((73, 5), (906, 3))
            >>> positive, negative = OpticalNeighbours.get_two_class_opt(frame, target_support, 
            ...                                                          max_distance=15, prob=0.20)
            >>> positive.shape, negative.shape
            ((116, 5), (814, 3))
        """
        corr = correlate(frame, support, max_distance, ('ra', 'dec'), ('ra', 'dec'),
                         add_distance=True, lsuffix='', rsuffix='_support')

        OpticalNeighbours._add_clear_distance(corr, 'id_support', max_distance, prob)

        positive = OpticalNeighbours._get_positive(corr)
        negative = OpticalNeighbours._get_negative(corr)

        return positive, negative


# ----------------------------------------------------------------------------------------------- #
#                                   PrepareClasses class                                          #
# ----------------------------------------------------------------------------------------------- #

class PrepareClasses():
    """
    Класс для объединения отобранных в модуле OpticalNeighbours оптических компаньонов с основным \
    рентгеновским каталогом для получения пар положительного (являются верным отождествлением) \
    и отрицательного (не являются верным отождествлением) класса.

    Для использования рекомендуется вызывать метод get_dataset() от результата метода get_pair() класса OpticalNeighbours.

    На вход подается два множества - верные оптические компаньоны (оптические объекты, которые излучают в рентгене) \
    и неверные оптические компаньоны (не изучают в рентгене).

    Процесс объединения этих двух множеств с рентгеновским каталогом производится по следующему алгоритму:

    * Находятся пары рентген-оптика для оптических объектов обоих классов так, что дистанция в парах менее 30 угловых секунды.

    * Удаляются пары с верными оптическими компаньонами, расстояния между рентгеновскими источниками из основного \
        и вспомогательного каталога менее r98.

    * Удаляются пары с верными оптическими компаньонами, в которых встречается неоднозначность соотнесения \
        (один к многим и многие к одному).

    * Среди пар с неверными оптическими источниками остаются только те, у которых рентгеновский источник имеет \
        верного оптического компаньона в наборе.

    Полученные два класса и являются положительным и отрицательным классом.
    """

    @staticmethod
    def _drop_pairs_with_diff_target(data: pd.DataFrame) -> None:
        """
        Оставляет только пары с расстоянием между источниками из основного и вспомогательного каталога менее r98.

        Функция изменяет входную таблицу.

        Таким образом полагается, что оптический компаньон рентгеновского источника из вспомогательного каталога \
            также является и компаньоном рентгеновского источника из основного каталога.

        Args:
            data: Каталог с парами рентген(основной)-оптика(вспомогательный) из положительного класса, \
                но с присутсвующими координатами рентгеновских источников вспомогательного каталога, \
                которые использовались для отбора верных оптических каталогов.\
                    Columns: id_target, ra_target, dec_target, poserr, ra_support, dec_support
        Note:
            Функция изменяет каталог - удаляет некоторые строки. Remove columns: ra_support, dec_support
        """
        add_range(data, ('ra_target', 'dec_target'), ('ra_support', 'dec_support'))
        data['_r98'] = (-2 * (data["poserr"]**2) * np.log(1 - 0.98))**0.5
        data.query('distance<=_r98', inplace=True)
        data.drop(columns=['ra_support', 'dec_support', 'distance', '_r98'], inplace=True)

    @staticmethod
    def _delete_unreliable(good_pair: pd.DataFrame) -> None:
        """
        Удаляет пары положительного класса с источниками, которые встречаются в нескольких парах - \
        то есть их компаньон неоднозначно определен. Аналогично удаляет и пары с оптическими источниками,\
        встречающимися в нескольких парах.

        Args:
            good_pair: Каталог с парами рентген(основной)-оптика(вспомогательный) из положительного класса.\
                Columns: 'id_target', 'id_frame'
        Note:
            Функция изменяет каталог - удаляет некоторые строки
        """
        OpticalNeighbours._delete_mult_obj(good_pair, 'id_target')
        OpticalNeighbours._delete_mult_obj(good_pair, 'id_frame')

    @staticmethod
    def _get_positive_pairs(frame_positive: pd.DataFrame, target: pd.DataFrame, max_distance: float) -> pd.DataFrame:
        """
        Получает позитивные пары отождествлений frame-target, отмеченные меткой класса 1.

        Args:
            frame_positive: Каталог frame источников, которые были определены как верные компаньоны target источников.\
                Columns: id, ra, dec, poserr, ra_support, dec_support
            target: Рентгеновский основной каталог. Columns: id, ra, dec, poserr
            max_distance: Максимальная дистанция в парах.
        Return:
            Совмещенный каталог рентген-оптика, с помеченнами парами 1 класса.\
                Columns: id_frame, ra_frame, dec_frame, id_target, ra_target, dec_target, poserr, mark
        """
        positive_pairs = correlate(frame_positive, target, max_distance, ('ra', 'dec'), ('ra', 'dec'),
                                   add_distance=True, lsuffix='_frame', rsuffix='_target')
        positive_pairs.query('distance/poserr<=5', inplace=True)

        PrepareClasses._drop_pairs_with_diff_target(positive_pairs)
        PrepareClasses._delete_unreliable(positive_pairs)

        positive_pairs['mark'] = 1
        return positive_pairs

    @staticmethod
    def _get_negative_pairs(frame_negative: pd.DataFrame, target: pd.DataFrame, positive_id: np.ndarray, max_distance: float) -> pd.DataFrame:
        """
        Получает негативные пары отождествлений рентген-оптика, отмеченные меткой класса 0.

        Args:
            opt_negative: Каталог оптических источников, которые были определены как неверные компаньоны рентгеновских источников.
            xray: Рентгеновский основной каталог.
            positive_id: id рентгеновских источников, которые должны содержаться в выходном каталоге.
            max_distance: Максимальная дистанция в парах.
        Return:
            Совмещенный каталог рентген-оптика, с помеченнами парами 0 класса.
        """
        sources = target[target['id'].isin(positive_id)]
        negative_pairs = correlate(frame_negative, sources, max_distance, ('ra', 'dec'), ('ra', 'dec'),
                                   lsuffix='_frame', rsuffix='_target')

        negative_pairs['mark'] = 0
        return negative_pairs

    @staticmethod
    def get_two_class_pair(target: pd.DataFrame, frame_positive: pd.DataFrame, frame_negative: pd.DataFrame, max_distance: float = 15) -> pd.DataFrame:
        """
        Объединение каталогов оптических объектов с основным каталогом рентгеновских источников.

        Args:
            opt_positive: Каталог с парами рентген(основной)-оптика(вспомогательный) из положительного класса. \
                Каталог должен содержать позиции рентгеновских источников из вспомогатьельного каталога, \
                которые являются компаньонами оптических объектов.
            opt_negative: Каталог с парами рентген(основной)-оптика(вспомогательный) из положительного класса.
            max_distance: Максимальная дистанция которая может быть между парами.

        Returns:
            Каталог пар рентген-оптика, с меткой класса (класс 1 - верное отождествление, класс 0 - неверное).

        Examples:
            TODO
        """
        positive_pairs = PrepareClasses._get_positive_pairs(frame_positive, target, max_distance=max_distance)
        positive_id = np.array(positive_pairs['id_target'].unique())
        negative_pairs = PrepareClasses._get_negative_pairs(frame_negative, target, positive_id, max_distance=max_distance)

        data = pd.concat([positive_pairs, negative_pairs]).reset_index(drop=True)
        return data

    @staticmethod
    def get_marked_data(frame: pd.DataFrame, target: pd.DataFrame, target_support: pd.DataFrame,
                        max_distance: float = 15, probability: float = 0.05) -> pd.DataFrame:
        """Возвращает всевозможные пары объектов рентген-оптика в радиусе , где для каждой пары указана метка класса - верная пара(1) и неверная пара(0)."""
        positive, negative = OpticalNeighbours.get_two_class_opt(frame, target_support, max_distance, probability)
        return PrepareClasses.get_two_class_pair(target, positive, negative, max_distance)


# ----------------------------------------------------------------------------------------------- #
#                                  SampleSelection class                                          #
# ----------------------------------------------------------------------------------------------- #
    
class SampleSelection:
    """
    Класс для разделения обучающей выборки на тренеровочную и тестовую.

    Под главным потоком понимается сумма потоков в фильтрах, которые важны для исследования.
    Названия колонок с этими потоками задаются в файле конфигурации.
    Главный поток является основным критерием для разделения выборки на тренировочную и тестовую части.

    Разделение может производиться двумя способами:

    * base_split: разделение производится независимо от суммы потоков

    * improved_split: разделение производится таким образом, чтобы в тестовой выборке распределение главного потока \
        у источников соответствовало распределению главного потока у источников из всего рентгеновского каталога.

    Для использования improved_split необходимо задать поля потоков в нужных фильтрах в конфигурационном файле.
    Эти поля будут учитываться при разделении.

    Таким образом, без заданных фильтров можно использовать только base_split.
    """

    @staticmethod
    def _get_filter_columns(df: pd.DataFrame) -> list[str]:
        """Возвращает названия колонок с потоками в нужных фильтрах."""
        return list(filter(lambda x: x.startswith('filter'), df.columns))

    @staticmethod
    def _add_filter(df: pd.DataFrame, delete_filtered_columns: bool = True, delete_zero_flux_rows: bool = True) -> None:
        """
        Добавляет в каталог поле 'filter', как сумму потоков в фильтрах, заданных в файле конфига.

        Args:
            df: Каталог рентгеновских источников, в котором есть колонки с потоками в нужных фильтрах.
            delete_filtered_columns: Нужно ли удалять колонки с потоками в нужных фильтрах.
            delete_zero_flux_rows: Нужно ли удалять строки с нулевым главным потоком.
        Note:
            Функция изменяет каталог - удаляет некоторые строки и добавляет столбец 'filter' - главный поток.
        """
        df['filter'] = 0
        for i in SampleSelection._get_filter_columns(df):  # Добавить возможность фильтр не только потоком делать
            df['filter'] += df[i]
            if delete_filtered_columns:
                del df[i]
        if delete_zero_flux_rows:
            df.query('filter>0', inplace=True)  # Возможно стоит не удалять строки а заменять на медиану.
        else:
            df.loc[df['filter'] == 0, 'filter'] = df.loc[df['filter'] != 0, 'filter'].values[:len(df.loc[df['filter'] == 0])]
        df['filter'] = np.log10(df['filter'])

    @staticmethod
    def _prepare_filter_table(filter_table: pd.DataFrame) -> pd.DataFrame:
        """Загружает основной рентгеновский каталог с колонками id, ra, dec, r0. Также добавляет поле фильтр в этот каталог."""
        col = SampleSelection._get_filter_columns(filter_table) + get_attrs_columns(filter_table, ['id', 'ra', 'dec'])
        filter_table = filter_table[col]
        SampleSelection._add_filter(filter_table)
        return filter_table

    @staticmethod
    def _add_hostless_base(df: pd.DataFrame, share: float = 0.8, seed: int = 42) -> None:
        """Добавляет бездомных источников."""
        id_col = df.attrs['id_target']
        frequent_ids = df.groupby(id_col).count().query('mark>=2').index
        positive = df.query(f'mark==1 and {id_col} in @frequent_ids')
        positive_id = positive.sample(n=int(len(positive) * (1-share)), replace=False)[id_col]
        print(len(positive_id))
        df.query(f'mark==0 or {id_col} not in @positive_id', inplace=True)

    @staticmethod
    def _add_hostless_improved(df: pd.DataFrame, share: float = 0.8, seed: int = 42) -> None:
        """Добавляет бездомные источники на основании статистики по всему каталогу."""
        warnings.warn("Еще не рализованно. Вызовется обычный добавление бездомных источников.")
        SampleSelection._add_hostless_base(df, share)

    @staticmethod
    def add_hostless(df: pd.DataFrame, method: str = 'base', share: float = 0.8, seed: int = 42, 
                     df_to_filter: Union[pd.DataFrame, None] = None) -> None:
        """
        Добавляет бездомные объекты в каталог, позволяя при этом получать метрики классификации бездомных источников. Без добавления\
        бездомных источников метрики не посчитаются. (в наборе не будет бездомных источников для классификации).

        Пока реализовано добавление 'base' - удаление (1-share)% верных пар, таким образом получаем (1- share)% бездомных источников.

        Будет реализованно 'improved' - попытка построить бездомные на основании основного каталога.
        Для использования improved method необходимо установить параметр df_to_filter, в котором будет содержаться \
            целевой поток, использование которого будет приводить к тому, что бездомные источники будут добавлены в зависимости от этого потока.
        Целевой поток достаточно критично влияет на вероятность быть бездомным на реалиных данных, так что использование этого метода позволит\
            получить тестовую выборку гораздо более репрезентативную.

        Args:
            df: Каталог пар рентген-оптика.
            method: Тип добавления бездомных источников. Есть варианты ['no','base','improved'].
            share: Доля оставшихся объектов с парами.
            seed: np.random.seed().
            df_to_filter: Каталог с целевыми потоками.
        
        Note:
            Функция изменяет каталог - удаляет некоторые строки.

        Examples:
            Базовый пример добавления бездомных источников в каталог. В общем случае не рукомендуется использовать этот модуль непосредственно,\
                лучше исполнять его в рамках :class:`cosmatch.match.Pipeline`.

            >>> import pandas as pd
            >>> import numpy as np
            >>> from cosmatch.match.data_handler import SampleSelection
            >>> from cosmatch.utils import correlate
            >>> from cosmatch.fake import generate_catalog
            >>>
            >>> size_opt = 1000; size_xray = 100
            >>> 
            >>> opt = generate_catalog(name='opt', size=size_opt, random_state=42)
            >>> xray = generate_catalog(name='xray', size=size_xray, random_state=42)
            >>> df = correlate(opt, xray, 10, add_distance=True)
            >>>
            >>> # Ставим метку верного отождествления для пар с минимальной дистанцией (для эксперимента)
            >>> min_distance = df.groupby('id_xray')['distance'].transform('min')
            >>> df['mark'] = df['distance'] == min_distance
            >>> df['mark'].sum() == df['id_xray'].nunique()
            True

            Для каждого уникального xray источника у нас есть верная с ним пара. Теперь уберем 20% таких пар:

            >>> SampleSelection.add_hostless(df, method='base', share=0.8, seed=42)
            >>> df['mark'].sum(), df['id_xray'].nunique()
            (65, 100)

            Важно заметить, что удаляются только пары с позитивными метками у тех target источников,\
                для которых в наборе данных есть еще одна пара с меткой mark=0 

            Метод "improved" в разработке TODO.
        """
        if method == 'base':
            SampleSelection._add_hostless_base(df, share, seed)
        elif method == 'improved':
            SampleSelection._add_hostless_improved(df, share, seed)
        elif method == 'no':
            return
        else:
            raise ValueError('Wrong "method" argument. Should be one of ["no","base","improved"].')
        

    @staticmethod
    def _base_split(df: pd.DataFrame, train_size: float = 0.8, seed: int = 42) -> Tuple[pd.DataFrame, pd.DataFrame, np.ndarray, np.ndarray]:
        """
        Разделяет случайно выборку на тренеровочную и тестовую часть (случайные рентгеновские источники берутся в train,\
        остальные в тест).

        Args:
            df: Каталог с парами рентген-оптика
            train_size: Доля рентгеновских источников, пары с которыми идут в тренеровочную выборку.
            seed: np.random.seed().
        Returns:
            - Тренеровочный набор данных без меток класса.
            - Тестовый набор данных без меток класса.
            - Метки класса для пар из тренеровочного наборы данных.
            - Метки класса для пар из тестового набора данных.
        """
        id_col = df.attrs['id_target']
        sources = np.array(df[id_col].unique())
        np.random.seed(seed)
        train_scs = set(np.random.choice(list(sources), int(len(sources) * train_size), replace=False))
        train = df[df[id_col].isin(train_scs)]
        test_scs = set(sources) - set(train_scs)
        test = df[df[id_col].isin(test_scs)]
        return train.drop(columns=['mark']), test.drop(columns=['mark']), \
            np.array(train['mark']), np.array(test['mark'])

    @staticmethod
    def _check_df_to_filter(df_to_filter: pd.DataFrame) -> None:
        """Проверяет, что все колонки с потоками в нужных фильтрах."""
        columns = SampleSelection.get_filter_columns(df_to_filter)
        if len(columns) == 0:
            raise ValueError('Не найдены колонки с потоками в нужных фильтрах.')
        check_attrs_structure(df_to_filter, ['id', 'ra', 'dec'])

    @staticmethod
    def _improved_split(df: pd.DataFrame, df_to_filter: pd.DataFrame, train_size: float = 0.8,
                        seed: int = 42) -> Tuple[pd.DataFrame, pd.DataFrame, np.ndarray, np.ndarray]:
        """
        Разделяет выборку на тренеровочную и тестовую часть в зависимости от главного потока источников\
        Распределение главного потока в тестовой выборке будет соответствовать распределению во всем основном рентгеновском каталоге.

        Args:
            train_size: Доля рентгеновских источников, пары с которыми идут в тренеровочную выборку.
            seed: np.random.seed().
        Returns:
            - Тренеровочный набор данных без меток класса.
            - Тестовый набор данных без меток класса.
            - Метки класса для пар из тренеровочного наборы данных.
            - Метки класса для пар из тестового набора данных.
        """
        SampleSelection._check_df_to_filter(df_to_filter)
        id_col = df.attrs['id_target']
        sources = df[id_col].unique()
        filter_table = SampleSelection._prepare_filter_table(df_to_filter)
        dist, borders = np.histogram(filter_table["filter"], bins=30)
        dist = (dist / dist.sum() * len(sources) * (1 - train_size)).astype(int)

        SampleSelection._add_filter(df, False)
        group = df[[id_col, 'filter']].drop_duplicates()
        del df['filter']

        test = group.head(1)
        for i in range(len(dist)):
            if dist[i] > 0:
                bot = borders[i]
                top = borders[i + 1]
                q = group.query('filter>=@bot and filter<@top')
                if len(q):
                    sample = q.sample(dist[i], replace=True, random_state=seed)
                test = pd.concat([test, sample])
        test = np.array(test[id_col])
        X_test = df[df[id_col].isin(test)]
        X_train = df[df[id_col].isin(test)]
        y_test = np.array(X_test['mark'])
        y_train = np.array(X_train['mark'])
        del X_test['mark'], X_train['mark']

        return X_train, X_test, y_train, y_test

    @staticmethod
    def get_splited_data(df: pd.DataFrame, split_method: str = 'base', hostless_method: str = 'no', train_size: float = 0.8, seed: int = 42, 
                         df_to_filter: Union[pd.DataFrame, None] = None) -> Tuple[pd.DataFrame, pd.DataFrame, np.ndarray, np.ndarray]:
        """
        Разделяет входной каталог на обучающую и тестовую выборку определенным способом.

        Args:
            df: Каталог верных и неверный пар рентген-оптика
            split_method: Вариант разбиения - ['base', 'impoved']
            hostless_method: Вариант добавления бездомных источников - ['no','base','improved']
            train_size: Пропорция тренеровочной выборки.
            seed: np.random.seed().
            df_to_filter: Каталог для определения статистики по основному потоку.

        Returns:
            - Тренеровочный набор данных без меток класса.
            - Тестовый набор данных без меток класса.
            - Метки класса для пар из тренеровочного наборы данных.
            - Метки класса для пар из тестового набора данных.

        Example:
            >>> from cosmatch.fake import generate_correlated_catalog
            >>> from cosmatch.match.data_handler import SampleSelection
            >>> 
            >>> ## Создаем совмещенный каталог
            >>> data = generate_correlated_catalog(name_frame='frame', name_target='target', 
            ...                                    size_frame=500, size_target=200,
            ...                                    random_state=42, num_frame_flux=0, 
            ...                                    num_target_flux=0,max_distance=10, distance=True, 
            ...                                    add_mark=True, mark_method='nearest')
            >>> 
            >>> data.mark.value_counts()
            >>> # mark
            >>> # 0    26021
            >>> # 1      200
            >>> # Name: count, dtype: int64
            >>> 
            >>> train, test, y_train, y_test = SampleSelection.get_splited_data(data, split_method='base', hostless_method='no',
            ...                                                                  train_size=0.8, seed=42)
            >>> y_train.sum(), y_test.sum()
            >>> # (160, 40)
            >>> 
            >>> len(train), len(test)
            >>> # (21070, 5151)

            Рассмотрим добавление бездомных объектов:

            >>> train, test, y_train, y_test = SampleSelection.get_splited_data(data, split_method='base', hostless_method='base',
            ...                                                                  train_size=0.8, seed=42)
            >>> y_train.sum(), y_test.sum(), 
            >>> # (125, 36)
            >>> len(train), len(test)
            >>> # (21035, 5147)
        """
        data = df.copy()
        SampleSelection.add_hostless(data, method=hostless_method, df_to_filter=df_to_filter)

        if split_method == 'base':
            return SampleSelection._base_split(data, train_size=train_size, seed=seed)
        elif split_method == 'improved':
            if df_to_filter is None:
                raise ValueError('Не был задан каталог для определения статистики по основному потоку. df_to_filter должен быть задан.')
            return SampleSelection._improved_split(data, train_size=train_size, seed=seed, df_to_filter=df_to_filter)
        else:
            warnings.warn('Incorrect type of spliting train data. Base was chosen.')
            return SampleSelection._base_split(data, train_size=train_size, seed=seed)


# ----------------------------------------------------------------------------------------------- #
#                              Prepare train/predict dataset                                      #
# ----------------------------------------------------------------------------------------------- #

def _make_order_in_columns(df: pd.DataFrame) -> pd.DataFrame:
    """Сортирует колонки в df."""
    main_col = [df.attrs['id_frame'], df.attrs['id_target'], 'mark',
                df.attrs['ra_frame'], df.attrs['dec_frame'], df.attrs['ra_target'], df.attrs['dec_target']]
    other_col = set(df.columns) - set(main_col)
    frame_col = sorted(list(filter(lambda x: x.startswith(df.attrs['name_frame'] + '_'), other_col)))
    target_col = sorted(list(filter(lambda x: not x.startswith(df.attrs['name_frame'] + '_'), other_col)))
    columns = main_col + frame_col + target_col
    if 'mark' not in df.columns:
        columns.remove('mark')
    return df[columns]


def _join_tables_with_jointable(join_table: pd.DataFrame, frame: pd.DataFrame, target: pd.DataFrame) -> pd.DataFrame:
    """Корректно джойнит таблицы, добавляя при этом attrs и изменяя имена колонок."""
    frame_main_col = [frame.attrs['id'], frame.attrs['ra'], frame.attrs['dec']]
    target_main_col = [target.attrs['id'], target.attrs['ra'], target.attrs['dec'], target.attrs['poserr']]

    rename_frame = {col: frame.attrs['name'] + '_' + col if col not in frame_main_col else col for col in frame.columns}
    rename_target = {col: target.attrs['name'] + '_' + col if col not in target_main_col else col for col in target.columns}

    data = join_table.join(frame.rename(columns=rename_frame).set_index(frame.attrs['id']), on='id_frame', how='left')
    data = data.join(target.rename(columns=rename_target).set_index(target.attrs['id']), on='id_target', how='left')
    data.rename(columns={'id_frame': frame.attrs['id'], 'id_target': target.attrs['id'],
                         'ra_frame': frame.attrs['ra'], 'dec_frame': frame.attrs['dec'],
                         'ra_target': target.attrs['ra'], 'dec_target': target.attrs['dec']}, inplace=True)
    
    unite_attrs_from_frame_and_target(data, frame, target)

    return data


def _prepare_proper_copies(df: pd.DataFrame, columns: list[str] = ['id', 'ra', 'dec']) -> pd.DataFrame:
    """Подгатавливает копии для построения основы будущего совмещенного каталога, к которому потом будут присоединены оставшиеся колонки."""
    check_attrs_structure(df, columns)
    main_columns = [df.attrs[col] for col in columns]
    df_tmp = df[main_columns].copy()
    rename = {df.attrs[col]: col for col in columns}
    df_tmp.rename(columns=rename, inplace=True)
    return df_tmp


def get_pairs_train(frame: pd.DataFrame, target: pd.DataFrame, target_support: pd.DataFrame,
                        max_distance: float = 15, probability: float = 0.05) -> pd.DataFrame:
    """
    Возвращает маркированный каталог пар рентген-оптика с необработанными колонками.
    
    Args:
        frame: Каталог с frame объектами. Attrs: name, id, ra, dec
        target: Каталог с target объектами. Attrs: name, id, ra, dec, poserr
        target_support: Каталог с target_support объектами. Attrs: name, id, ra, dec
        max_distance: Максимальное расстояние в угловых секундах между frame и target_support объектами, которое будет в полученных парах. Следует выбирать в зависимости от\
            того, как были получены наборы данных. Не рекомендуется делать менее 10.
        probability: Вероятность, что frame объект внутри радиуса будет неверным компаньоном для target источника.\
            Напрямую влияет на размер положительной и отрицательной выборки. 

    Returns:
        Набор пар frame-target со всеми колонками из обоих наборов данных, включая колонку mark. \
            mark=1 - верное отождествление, mark=0 - неверное отождествление.

    Examples:
        Базовый пример работы функции:

        >>> from cosmatch.match import get_pairs_train
        >>> from cosmatch.fake import generate_catalog
        >>>
        >>> frame = generate_catalog(name='frame', size=100, random_state=42)
        >>> frame.columns
        >>> # Index(['id_frame', 'ra_frame', 'dec_frame'], dtype='object')
        >>> frame.attrs
        >>> # {'name': 'frame', 'id': 'id_frame', 'ra': 'ra_frame', 'dec': 'dec_frame'}
        >>> 
        >>> target = generate_catalog(name='target', size=10, random_state=42, other_columns=['poserr'])
        >>> target.attrs['poserr'] = 'poserr'
        >>>
        >>> target_support = generate_catalog(name='support', size=10, random_state=42)
        >>>
        >>> result = get_pairs_train(frame, target, target_support, max_distance=10, probability=0.05)
        >>> result.columns
        >>> # Index(['id_frame', 'id_target', 'mark', 'ra_frame', 'dec_frame', 'ra_target', 'dec_target', 'poserr'], dtype='object')

        Вы наборах данных могут содержаться любые дополнительные колонки, после выполнения функции они будут добавлены в результирующий набор.\
            Ко всем колонкам, кроме основных ('id', 'ra', 'dec', 'poserr') будет добавлен постфикс f'_{name}', где name соответствующий атрибут таблицы.

        >>> frame['flux_1'] = ...
        >>> frame['flux_2'] = ...
        >>> target['flux_1'] = ...
        >>> target['flux_2'] = ...
        >>> result = get_pairs_train(frame, target, target_support, max_distance=10, probability=0.05)
        >>> result.columns
        >>> # Index(['id_frame', 'id_target', 'mark', 'ra_frame', 'dec_frame', 'ra_target',
        >>> #        'dec_target', 'frame_flux_1', 'frame_flux_2', 'poserr', 'target_flux_1',
        >>> #        'target_flux_2'], dtype='object')

        Максимальная дистанция в совмещенном каталоге будет меньше заданного порога в max_distance=10 угловых секунд:

        >>> from cosmatch.utils import add_range
        >>> add_range(result)
        >>> result['distance'].max()
        >>> # 9.895813043836682

        В результирующем наборе будут следующие атрибуты:

        >>> result.attrs
        >>> # {'ids': ['id_frame', 'id_target'],
        >>> # 'coords': ['ra_frame', 'dec_frame', 'ra_target', 'dec_target'],
        >>> # 'name_frame': 'frame', 'id_frame': 'id_frame', 'ra_frame': 'ra_frame', 'dec_frame': 'dec_frame',
        >>> # 'name_target': 'target', 'id_target': 'id_target', 'ra_target': 'ra_target', 'dec_target': 'dec_target'}

        В дальнейшей работе программы эти атрибуты будут использоваться для добавления новых признаков, игнорирования признаков и многих других\
            действий. Не рекомендуется изменять имена атрибутов без крайней необходимости.
    """
    frame_tmp = _prepare_proper_copies(frame, ['id', 'ra', 'dec'])
    target_tmp = _prepare_proper_copies(target, ['id', 'ra', 'dec', 'poserr'])
    target_support_tmp = _prepare_proper_copies(target_support, ['id', 'ra', 'dec'])

    join_table = PrepareClasses.get_marked_data(frame_tmp,
                                                target_tmp,
                                                target_support_tmp,
                                                max_distance=max_distance,
                                                probability=probability)
    join_table = join_table[['id_frame', 'id_target', 'mark']]

    data = _join_tables_with_jointable(join_table, frame, target)
    data = _make_order_in_columns(data)
    return data


def get_pairs_predict(frame: pd.DataFrame, target: pd.DataFrame, max_distance: float = 15) -> pd.DataFrame:
    """
    Возвращает каталог всевозможных пар рентген-оптика в заданном радиусе с необработанными колонками.

    Данный каталог можно применять для предсказания верных отождествлений - по сути на построенном множестве пар и решается задача отождествления.

    Args:
        frame: каталог c frame объектами. Attrs: ['name', 'id', 'ra', 'dec']
        target: каталог c target объектами. Attrs: ['name', 'id', 'ra', 'dec', 'poserr']
        max_distance: максимальное расстояние в парах между объектами объектами в угловых секундах.

    Returns:
        Набор пар frame-target со всеми колонками из обоих наборов данных. Attrs:\
            ['name_frame', 'name_target', 'id_frame', 'id_target', 'ra_frame', 'dec_frame', 'ra_target', 'dec_target', 'ids', 'coords']

    Examples:
        Пример работы на сгенерированном каталоге:

        >>> from cosmatch.match import get_pairs_predict
        >>> from cosmatch.fake import generate_catalog
        >>> 
        >>> frame_size = 200; target_size = 100
        >>> 
        >>> frame = generate_catalog('opt', size=frame_size, random_state=42)
        >>> 
        >>> target = generate_catalog('xray', size=target_size, random_state=42, other_columns=['poserr'])
        >>> target.attrs['poserr'] = 'poserr'
        >>>
        >>> pairs = get_pairs_predict(frame, target, max_distance=15)
        >>> pairs.columns
        >>> # Index(['id_opt', 'id_xray', 'ra_opt', 'dec_opt', 'ra_xray', 'dec_xray', 'poserr'], dtype='object')
        >>> pairs.shape
        >>> # (9725, 7)

        Дистанция между парами будет меньше заданного порога в max_distance=15 угловых секунд:

        >>> from cosmatch.utils import add_range
        >>> add_range(pairs)
        >>> pairs['distance'].max()
        >>> # 14.998417402192134
    """
    frame_tmp = _prepare_proper_copies(frame, ['id', 'ra', 'dec'])
    target_tmp = _prepare_proper_copies(target, ['id', 'ra', 'dec'])

    join_table = correlate(frame_tmp, target_tmp, max_distance, ('ra', 'dec'), ('ra', 'dec'), lsuffix='_frame', rsuffix='_target')
    join_table = join_table[['id_frame', 'id_target']]
    data = _join_tables_with_jointable(join_table, frame, target)
    data = _make_order_in_columns(data)
    return data
